from typing import List, Dict, Any, Set, Tuple, Union
from flitsr.spectrum import Spectrum, Outcome


class SpectrumBuilder:
    """ Builds a `Spectrum <flitsr.spectrum.Spectrum>` object."""

    def __init__(self, collapse_methods: bool = False):
        """
        Constructs a `SpectrumBuilder` object to facilitate building a
        `Spectrum <flitsr.spectrum.Spectrum>`.

        Args:
          collapse_methods: bool: (Default value = False) Whether to form a
            method level spectrum.
        """
        self._collapse_methods = collapse_methods
        self._method_map: Dict[int, Spectrum.Element] = {}
        self._methods: Dict[Tuple[str, str], Spectrum.Element] = {}
        self._elements: List[Spectrum.Element] = []
        self._tests: Dict[int, Spectrum.Test] = {}
        self._executions: Dict[Spectrum.Test, Set[Spectrum.Element]] = {}

    def get_tests(self) -> List[Spectrum.Test]:
        """ Return the current list of Tests. """
        return [self._tests[key] for key in sorted(self._tests.keys())]

    def getElement(self, index: int):
        """
        Retreive the element with the specified index.

        Args:
          index: int: The index of the element to retrieve.

        Returns:
          The `Element <flitsr.spectrum.Spectrum.Element>` corresponding to the
          `index` given.

        Raises:
          KeyError: If the `index` does not correspond to an element.
        """
        return self._method_map[index]

    def addTest(self, name: str, outcome: Outcome, index: int = None):
        """
        Add a new test to the spectrum with the given name, index, and outcome.

        Args:
          name: The name of the test.
          index: (Default value = None) A unique identifier for this test.
          outcome: The outcome (pass/fail) of the test.
        """
        if (index is None):
            index = len(self._tests)
        t = Spectrum.Test(name, index, outcome)
        self._tests[index] = t
        self._executions[t] = set()
        return t

    def addElement(self, details: List[str],
                   faults: List[Any], index: int = None) -> Spectrum.Element:
        """
        Add a new element to the spectrum, with the given details and faults.

        Args:
          details: List[str]: The list of details of the element. The list MUST
            contain at least one string. It conforms to one of the following::

              [<name>]
              [<path>, <method>]
              [<path>, <line num>]
              [<path>, <method>, <line num>]

          faults: List[Any]: The list of faults that this element pertains to.
          index: int:  (Default value = None) The (optional) index of this
            element.

        Returns:
          The created element.
        """
        if (index is None):
            index = len(self._elements)
        i = len(self._method_map)  # number of actual elements
        if (self._collapse_methods):
            if ((details[0], details[1]) not in self._methods):
                # append with first line number
                elem = self._add_element(details, faults, index)
                self._methods[(details[0], details[1])] = elem
                self._method_map[i] = elem
            else:
                elem = self._method_map[i] = self._methods[(details[0],
                                                            details[1])]
                # update method's faults
                for fault in faults:
                    if (fault not in elem.faults):
                        elem.faults.append(fault)
        else:
            elem = self._add_element(details, faults, index)
            self._method_map[i] = elem
        return elem

    def _add_element(self, details: List[str], faults: List[Any],
                     index: int) -> Spectrum.Element:
        """
        Helper method to consistently add element to this builder.

        Args:
          details: List[str]:
          faults: List[Any]:
          index: int:

        Returns:

        """
        # Create element
        e = Spectrum.Element(details, index, faults)
        # Record element
        self._elements.append(e)
        return e

    def addExecution(self, test: Union[Spectrum.Test, int],
                     elem: Union[Spectrum.Element, int]):
        """
        Mark the specified element as executed in the given test.

        Args:
          test: Union[Spectrum.Test, int]: The test to add the execution to.
            Can be either a `Test <flitsr.spectrum.Spectrum.Test>` object OR an
            integer representing a valid index of a Test.
          elem: Union[Spectrum.Element, int]: The element to add the execution
            for. Can be either an `Element <flitsr.spectrum.Spectrum.Element>`
            object OR an integer representing a valid Element index.

        Raises:
          TestKeyError: When `test` is an int and does not refer to a valid
            test index.
          ElemKeyError: When `elem` is an int and does not refer to a valid
            element index.
        """
        # Get the test or raise execption
        if (isinstance(test, int)):
            try:
                test = self._tests[test]
            except KeyError:
                raise TestKeyError()

        # Get the element or raise execption
        if (isinstance(elem, int)):
            try:
                elem = self._method_map[elem]
            except KeyError:
                raise ElemKeyError()

        # add execution
        if (test not in self._executions):
            self._executions[test] = set()
        self._executions[test].add(elem)

    def addNonExecution(self, test: Spectrum.Test, elem: Spectrum.Element):
        """
        Mark the specified element as not executed in the given test.
        NOTE: Adding non-executions is not required for the spectrum
        (non-execution is assumed by default), thus this method does not do
        anything.

        .. deprecated:: 2.2
           There is no need to call this function anymore as non-executions are
           not stored by default.

        Args:
          test: Spectrum.Test: The Test to add the non-execution to.
          elem: Spectrum.Element: The Element to add to non-execution of.
        """
        # if (test not in self._executions):
        #     self._executions[test] = {}
        # self._executions[test][elem] = False
        pass

    def form_groups(self) -> List[Spectrum.Group]:
        """
        Given one test pertaining to a row in the spectrum, split the groups
        according to the coverage. The split will potentially split each group
        in half, based on those elements that are executed in the test, and
        those that are not.
        """
        stack: List[Tuple[Set[Spectrum.Element], int]] = []
        groups: List[Spectrum.Group] = []
        test_list = self.get_tests()
        # Initialize stack with all elements in one group
        stack.append((set(self._elements), 0))
        while (len(stack) > 0):
            cur, t_ind = stack.pop()
            if (t_ind >= len(test_list) or len(cur) == 1):
                groups.append(Spectrum.Group(list(cur)))
            else:
                test = test_list[t_ind]
                exe_elem = cur.intersection(self._executions[test])
                nexe_elem = cur.difference(exe_elem)
                if (len(exe_elem) != 0):
                    stack.append((exe_elem, t_ind+1))
                if (len(nexe_elem) != 0):
                    stack.append((nexe_elem, t_ind+1))
        return groups

    def get_spectrum(self) -> Spectrum:
        """ Return the spectrum from this `SpectrumBuilder`. """
        groups = self.form_groups()
        spectrum = Spectrum(self._elements, groups, self.get_tests(),
                            self._executions)
        return spectrum


class TestKeyError(KeyError):
    """
    Exception that is raised when a Test is not found in the `SpectrumBuilder`.
    """
    pass


class ElemKeyError(KeyError):
    """
    Exception that is raised when an Element is not found in the
    `SpectrumBuilder`.
    """
    pass


class SpectrumUpdater(SpectrumBuilder):
    """
    Allows updating of an already created spectrum using the `SpectrumBuilder`
    interface.
    """
    def __init__(self, spectrum: Spectrum):
        super().__init__(collapse_methods=False)
        self._elements = spectrum._elements[:]
        self._tests = {t.index: t for t in spectrum.tests()[:]}
        self._executions = dict()
        # copy all the executions
        for test in self.get_tests():
            new_exe = self._executions.setdefault(test, set())
            for group in spectrum.groups():
                # Only copy over actually executed elements
                if (spectrum[test][group]):
                    for elem in group:
                        new_exe.add(elem)

    def remove_test_with_executions(self, test: Spectrum.Test):
        """
        Remove the given `test` from the spectrum, including any executions
        related to it.

        Args:
          test: Spectrum.Test: The test to remove.
        """
        del self._tests[test.index]
        del self._executions[test]

    def copy_test_and_execution(self, test: Spectrum.Test, spectrum: Spectrum):
        """
        Copy over the given `test` to this spectrum, with the executions given
        in the supplied other `spectrum`.

        Args:
          test: Spectrum.Test: The test to copy over to this spectrum.
          spectrum: Spectrum: The spectrum containing the executions of `test`.
        """
        self._tests[test.index] = test
        new_exe = self._executions.setdefault(test, set())
        for group in spectrum.groups():
            # Only copy over actually executed elements
            if (spectrum[test][group]):
                for other in group:
                    # get this spectrum's element
                    elem = self._get_element(other)
                    new_exe.add(elem)

    def _get_element(self, other: Spectrum.Element) -> Spectrum.Element:
        """
        Retrieve the element from this spectrum that is semantically equal to
        the `other` element.

        Args:
          other: Spectrum.Element: The element to find a semantic equivalent
            for.

        Returns:
          The element from this spectrum that is corresponds to the `other`
            element given.

        Raises:
          KeyError: If no element could be found that is semantically equal to
            the given `other` element.
        """
        for elem in self._elements:
            if (elem.semantic_eq(other)):
                return elem
        raise KeyError(f'Could not find element {other} in SpectrumUpdater')



import sys
import re
import os
from os import path as osp
from io import TextIOWrapper
from shutil import rmtree
from flitsr.split_faults import split_spectrum_faults, NoFaultsError
from flitsr.spectrum import Spectrum, Outcome
from flitsr.spectrumBuilder import SpectrumBuilder, TestKeyError, ElemKeyError
from flitsr.errors import error
from flitsr.input.input_reader import Input


class Gzoltar(Input):
    def get_run_file_name(self, input_path: str):
        return input_path.split("/")[0] + ".run"

    def construct_details(self, f: TextIOWrapper, method_level: bool,
                          sb: SpectrumBuilder):
        """
        Fills the spectrum object with elements read in from the open file 'f'.
        """
        bugs = 0
        assert f.readline() == 'name\n'  # remove header
        for line in f:
            m = re.fullmatch('([^$]+)\\$([^#]+)#([^:]+):([0-9]+)(?::(.+))?',
                             line.rstrip())
            if (m is None):
                error("Incorrectly formatted line \"" + line +
                      "\" when reading input file")
            else:
                faults = []
                if (m.group(5) is not None):
                    bs = m.group(5).split(':')
                    if (not bs[0].isdigit()):
                        faults = [bugs]
                    else:
                        faults = []
                        for b in bs:
                            faults.append(int(b))
                    bugs += 1
                details = [m.group(1)+"."+m.group(2), m.group(3), m.group(4)]
                sb.addElement(details, faults)

    def construct_tests(self, tests_reader: TextIOWrapper,
                        sb: SpectrumBuilder):
        tests_reader.readline()
        for line in tests_reader:
            m = re.fullmatch('([^,]+),(PASS|FAIL|ERROR)(,.*)?', line.rstrip())
            if (m is None):
                error("incorrectly formatted test line in input file:",
                      line, "terminating...")
            else:
                sb.addTest(m.group(1), Outcome[m.group(2)])

    def fill_spectrum(self, bin_file: TextIOWrapper, sb: SpectrumBuilder):
        for t, line in enumerate(bin_file):
            if (line == ''):
                error('Incorrect number of matrix lines', f'({t+1})',
                      'in input file, terminating...')
            arr = line.rstrip().split()
            # Loop over all elements in test exec (remove test outcome at end)
            for i in range(0, len(arr)-1):
                if (arr[i] != "0"):
                    try:
                        sb.addExecution(t, i)
                    except ElemKeyError:
                        error('Incorrect number of matrix columns',
                              f'({i+1})', 'in input file, terminating...')
                    except TestKeyError:
                        error('Incorrect number of matrix lines',
                              f'({t+1})', 'in input file, terminating...')

    def read_spectrum(self, input_path: str, split_faults: bool,
                      method_level=False) -> Spectrum:
        sb = SpectrumBuilder(method_level)
        # Getting the details of the elements
        self.construct_details(open(input_path+"/spectra.csv"),
                               method_level, sb)
        # Getting the details of the tests
        self.construct_tests(open(input_path+"/tests.csv"), sb)
        # Constructing the spectrum
        self.fill_spectrum(open(input_path+"/matrix.txt"), sb)
        spectrum = sb.get_spectrum()
        del sb
        # Split fault groups if necessary
        if (split_faults):
            try:
                split_spectrum_faults(spectrum)
            except NoFaultsError:
                error(f"No exposable faults in {input_path}, terminating...")
        return spectrum

    @staticmethod
    def check_format(input_path: str) -> bool:
        return (osp.isdir(input_path) and
                osp.isfile(osp.join(input_path, "matrix.txt")) and
                osp.isfile(osp.join(input_path, "spectra.csv")) and
                osp.isfile(osp.join(input_path, "tests.csv")))

    @classmethod
    def write_spectrum(cls, spectrum: Spectrum, directory: str):
        """ Output the spectrum in Gzoltar format """
        type_ = cls.get_type()
        # First check that the directory exists and is empty
        if (not osp.isdir(directory)):
            os.mkdir(directory)
        else:
            rmtree(directory)
            os.mkdir(directory)
        # Next print out the tests
        with open(osp.join(directory, "tests.csv"), 'w') as test_file:
            print("name,outcome,runtime,stacktrace", file=test_file)
            for test in spectrum.tests():
                print(test.name, "PASS" if test.outcome is Outcome.PASSED else
                      "FAIL", "", "", sep=",", file=test_file)
        with open(osp.join(directory, "spectra.csv"), 'w') as units_file:
            print("name", file=units_file)
            # TODO: change _elements below to elements()
            for elem in spectrum._elements:
                print(elem.output_str(type_=type_), file=units_file)
        with open(osp.join(directory, "matrix.txt"), 'w') as matrix_file:
            for test in spectrum.tests():
                for elem in spectrum._elements:
                    print(int(spectrum[test][elem]), end=" ", file=matrix_file)
                print('+' if test.outcome is Outcome.PASS else '-',
                      file=matrix_file)

    @classmethod
    def get_elem_separators(cls):
        return ['$', '#', ':', ':']


if __name__ == "__main__":
    from flitsr.output import print_spectrum
    d = sys.argv[1]
    ginput = Gzoltar()
    spectrum = ginput.read_spectrum(d, False)
    print_spectrum(spectrum)



from abc import ABC, abstractmethod
from flitsr.spectrum import Spectrum
from flitsr import advanced


class Refiner(ABC):
    """ A `Refiner` technique takes a `Spectrum <flitsr.spectrum.Spectrum>` and
    manipulates and extends it, returning a modified spectrum."""

    @abstractmethod
    def refine(self, spectrum: Spectrum, method_lvl=False) -> Spectrum:
        """
        Provides the functionality for the `Refiner`. Takes a `Spectrum
        <flitsr.spectrum.Spectrum>` and manipulates or extends it, returning a
        modified spectrum.

        Args:
          spectrum: Spectrum: The input `Spectrum <flitsr.spectrum.Spectrum>`
          method_lvl:  (Default value = False) Whether the input `spectrum`
            refers to a method level spectrum.

        Returns:
          The modified `Spectrum<flitsr.spectrum.Spectrum>`.

        """
        pass

    def __init_subclass__(cls):
        advanced.register_refiner(cls)



import sys
from flitsr.spectrum import Spectrum
from typing import Dict, List, Set, Tuple, Any
from flitsr.input.input_reader import Input
from flitsr.errors import warning


class NoFaultsError(ValueError):
    """
    Exception class for when splitting the fauls causes there to be no
    exposable faults left in the spectrum
    """
    pass


def split_spectrum_faults(spectrum: Spectrum):
    """
    Splits fault groups that are a combination of two or more sub-faults in
    mutually exclusive parts of the system into separate faults, as well as
    removing faults that are not exposed. Operates on the spectrum in-place.

    Raises:
        NoFaultsError: When there are no exposed faults in the spectrum after
            splitting. Note that the spectrum is still modified when this is
            thrown, so this can be ignored in the client code.
    """
    faults, unexposed = split(spectrum)
    for elem in unexposed:
        elem.faults.clear()
        warning(f"Dropped faulty UUT: {elem} due to unexposure")
    # Get element's fault lists
    fault_lists: Dict[Spectrum.Element, List[float]] = {}
    for (f_num, f_locs) in faults.items():
        for elem in f_locs:
            fault_lists.setdefault(elem, []).append(f_num)
    # Set element's fault lists
    for (elem, f_list) in fault_lists.items():
        elem.faults = f_list
    if (len(faults) == 0):
        raise NoFaultsError()


def split(spectrum: Spectrum) -> Tuple[Dict[float, Set[Spectrum.Element]],
                                       Set[Spectrum.Element]]:
    """
    Splits fault groups that are a combination of two or more sub-faults in
    mutually exclusive parts of the system into separate faults. Also finds
    faults that are not exposed and returns them separately.
    """
    faults = spectrum.get_faults()
    if (faults == {}):
        return {}, set()
    ftemp = [(set(elem), f[0], False) for f in faults.items() for elem in f[1]]
    for test in spectrum.failing():
        merge: Dict[Any, Set[Spectrum.Element]] = {}
        remain = []
        for equiv in ftemp:
            if (any(spectrum[test][e] for e in equiv[0])):
                merge.setdefault(equiv[1], set()).update(equiv[0])
            else:
                remain.append(equiv)
        if (len(merge) != 0):
            for (f_num, f_locs) in merge.items():
                remain.append((f_locs, f_num, True))
        ftemp = remain
    fmap: Dict[float, List[Set[Spectrum.Element]]] = {}
    unexposed: Set[Spectrum.Element] = set()
    for equiv in ftemp:
        if (not equiv[2]):
            unexposed.update(equiv[0])
            continue
        fmap.setdefault(equiv[1], []).append(equiv[0])
    new_faults: Dict[float, Set[Spectrum.Element]] = {}
    for item2 in fmap.items():
        if (len(item2[1]) == 1):
            new_faults[item2[0]] = item2[1][0]
        else:
            for i in range(len(item2[1])):
                new_faults[float("{}.{}".format(item2[0], i+1))] = item2[1][i]
    return new_faults, unexposed


if __name__ == "__main__":
    d = sys.argv[1]
    i = 2
    num_only = False
    while (True):
        if (len(sys.argv) > i):
            if (sys.argv[i] == "num"):
                num_only = True
            else:
                print("Unknown option:", sys.argv[i])
                quit()
            i += 1
        else:
            break
    spectrum = Input.read_in(d, False)
    faults = spectrum.get_faults()
    print("faults:", faults)
    # print(groups)
    # print_spectrum(spectrum)
    faults, unexposed = split(spectrum)
    print("split faults:", faults)
    print("unexposed:", unexposed)



from abc import ABC, abstractmethod
from typing import List
from flitsr.spectrum import Spectrum
from flitsr import advanced


class Cluster(ABC):
    """ A `Cluster` technique takes a `Spectrum <flitsr.spectrum.Spectrum>` and
    returns a collection of spectra that are decompositions of the original"""

    @abstractmethod
    def cluster(self, inp_file: str, spectrum: Spectrum,
                method_lvl=False) -> List[Spectrum]:
        """
        Provides the clustering functionality of this `Cluster` technique. A
        `Cluster` takes a `Spectrum <flitsr.spectrum.Spectrum>` `spectrum` and
        returns a collection of spectra which are decompositions of the input
        spectrum.

        Args:
          inp_file: str: The input file for the input spectrum
          spectrum: Spectrum: The input spectrum
          method_lvl:  (Default value = False) Whether the input spectrum
            refers to a method level spectrum

        Returns:
          A collection of spectra which are decompositions of the input
          spectrum.
        """
        pass

    def __init_subclass__(cls):
        advanced.register_cluster(cls)



from flitsr.tie import Ties, Tie
from flitsr.precision_recall import method


def one_top_n(ties: Ties, n: int, collapse=False) -> float:
    k, _ = method(n, ties, collapse=collapse)
    return min(1.0, k)


def all_top_n(ties: Ties, n: int, collapse=False) -> float:
    k, _ = method(n, ties, collapse=collapse)
    return k


def perc_top_n(ties: Ties, n: int, collapse=False) -> float:
    if (len(ties.faults) == 0):
        return 100
    else:
        k, _ = method(n, ties, collapse=collapse)
        return (k/len(ties.faults))*100


def size_top1(ties: Ties) -> int:
    tie = get_top1(ties)
    return len(tie.elems())


# <------------------------- Helper functions ---------------------->


def get_top1(ties: Ties) -> Tie:
    tie_iter = iter(ties)
    return next(tie_iter)



# PYTHON_ARGCOMPLETE_OK
import sys
import re
import copy
from os import path as osp
from math import log
from typing import List, Optional
from flitsr import weffort
from flitsr import top
from flitsr import percent_at_n
from flitsr import precision_recall
from flitsr.output import print_rankings, print_spectrum_csv
from flitsr import cutoff_points
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking, Rankings
from flitsr.tie import Ties
from flitsr.args import Args
from flitsr.advanced import ClusterType, RankerType
from flitsr.input.input_reader import Input
from flitsr.errors import error
from flitsr.advanced import Config


def compute_cutoff(cutoff: str, ranking: Ranking, spectrum: Spectrum,
                   mode: str, effort: str) -> Ranking:
    faults = spectrum.get_faults()
    if (cutoff.startswith("basis")):
        if ('=' in cutoff):
            num_bases = int(cutoff.split('=')[1])
        else:
            num_bases = 1
        ranking = cutoff_points.basis(num_bases, spectrum, faults, ranking,
                                      mode, effort=effort)
    else:
        ranking = cutoff_points.cut(cutoff, spectrum, faults, ranking, mode,
                                    effort=effort)
    return ranking


def output(rankings: Rankings, weff=[], top1=[],
           perc_at_n=[], prec_rec=[], faults=[], collapse=False,
           csv=False, decimals=2, file=sys.stdout):
    if (weff or top1 or perc_at_n or prec_rec or faults):
        ties: Ties = Ties(rankings)
        if (weff):
            if ("first" in weff):
                print("wasted effort (first): {:.{}f}".format(
                    weffort.first(ties, collapse),
                    decimals), file=file)
            if ("avg" in weff):
                print("wasted effort (avg): {:.{}f}".format(
                    weffort.average(ties, collapse),
                    decimals), file=file)
            if ("med" in weff):
                print("wasted effort (median): {:.{}f}".format(
                    weffort.median(ties, collapse),
                    decimals), file=file)
            if ("last" in weff):
                print("wasted effort (last): {:.{}f}".format(
                    weffort.last(ties, collapse),
                    decimals), file=file)
            for nth in [w for w in weff if type(w) == int]:
                print("wasted effort ({}): {:.{}f}".format(
                    nth, weffort.nth(ties, int(nth), collapse),
                    decimals), file=file)
        if (top1):
            for entry in top1:
                if (entry[0] == 'all'):
                    atp = top.all_top_n(ties, entry[1], collapse=collapse)
                    print(f"all top {entry[1]}: {atp:.{decimals}f}", file=file)
                elif (entry[0] == 'one'):
                    otp = top.one_top_n(ties, entry[1], collapse=collapse)
                    print(f"one top {entry[1]}: {otp:.{decimals}f}", file=file)
                elif (entry[0] == 'perc'):
                    ptp = top.perc_top_n(ties, entry[1], collapse=collapse)
                    print(f"perc top {entry[1]}: {ptp:.{decimals}f}",
                          file=file)
        if (perc_at_n):
            bumps = percent_at_n.getBumps(ties, collapse=collapse)
            if ('perc' in perc_at_n):
                form = ','.join(['{{:.{}f}}'.format(decimals)]*len(bumps))
                print("percentage at n:", form.format(*bumps), file=file)
            if (any([a in perc_at_n for a in ['auc', 'pauc', 'lauc']])):
                auc = percent_at_n.auc_calc(
                        percent_at_n.combine([(bumps[0], bumps[1:])]))
                if ('auc' in perc_at_n):
                    print("auc:", auc, file=file)
                if ('pauc' in perc_at_n):
                    optimal = percent_at_n.auc_calc([(0.0, 100.0)])
                    print("pauc:", "{:.{}f}".format(auc/optimal, decimals),
                          file=file)
                if ('lauc' in perc_at_n):
                    optimal = percent_at_n.auc_calc([(0.0, 100.0)])+1
                    lauc = abs(1-log(optimal-auc, optimal))
                    print("lauc:", "{:.{}f}".format(lauc, decimals), file=file)
        if (prec_rec):
            for entry in prec_rec:
                if (entry[0] == 'p'):
                    p = precision_recall.precision(entry[1], ties,
                                                   collapse=collapse)
                    print("precision at {}: {:.{}f}".format(entry[1], p,
                                                            decimals), file=file)
                elif (entry[0] == 'r'):
                    r = precision_recall.recall(entry[1], ties,
                                                collapse=collapse)
                    print("recall at {}: {:.{}f}".format(entry[1], r,
                                                         decimals), file=file)
        if (faults):
            if ("num" in faults):
                print("fault number: {}".format(len(ties.faults)), file=file)
            if ("ids" in faults):
                print("fault ids: {}".format(list(ties.faults.keys())),
                      file=file)
            if ("elems" in faults):
                print("fault elements: {}".format([e for es in ties.faults.values()
                                                   for e in es]), file=file)
            if ("all" in faults):
                print("fault info: {}".format(ties.faults), file=file)
    else:
        print_rankings(rankings, csv, file=file)


def main(argv: Optional[List[str]] = None):
    args: Args = Args(argv, cmd_line=True)
    # If only a ranking is given, print out metrics and return
    if (args.ranking):
        from flitsr.read_ranking import read_any_ranking
        rankings = read_any_ranking(args.input, method_level=args.method)
        output(rankings, args.weff, args.top1, args.perc_at_n,
               args.prec_rec, args.faults, args.collapse, csv=args.csv,
               decimals=args.decimals, file=args.output)
        return
    # Else, run the full process
    try:
        reader = Input.get_reader(args.input)
    except ValueError as e:
        error(e)
    d_p = reader.get_run_file_name(args.input)
    # Read the spectrum in and setup parallel if needed
    gspectrum = reader.read_spectrum(args.input, args.split, args.method)
    if (gspectrum is None or len(gspectrum.spectrum) == 0):
        print("ERROR: Incorrectly formatted input file, terminating...",
              file=sys.stderr)
        return
    if (args.spectrum_csv):
        print_spectrum_csv(gspectrum, file=args.output)
        return
    # Execute techniques
    config: Config
    for config in args.types:
        for metric in args.metrics:
            # Get the output channel
            if (len(args.metrics) == 1 and len(args.types) == 1 and not args.all):
                output_file = args.output
            else:
                # store output files in the current directory
                input_filename = osp.basename(d_p)
                filename = (config.get_file_name(args.print_params) + '_'
                            + metric + '_' + input_filename)
                try:
                    output_file = open(filename, "x")
                except FileExistsError:
                    if (args.no_override):
                        # print("WARNING: Skipping execution of already "
                        #       "existing file", filename, file=sys.stderr)
                        continue
                    else:
                        print("WARNING: overriding file", filename,
                              file=sys.stderr)
                        output_file = open(filename, 'w')
            # copy spectrum
            spectrum = copy.deepcopy(gspectrum)
            # Check for spectrum refining
            refiner = config.refiner(args)
            if (refiner is not None):
                spectrum = refiner.refine(spectrum, args.method)
            # Check for clustering
            cluster = config.cluster(args)
            # deal with clustering technique as metric
            if (cluster is None and hasattr(ClusterType, metric.upper())):
                metric_cluster = ClusterType[metric.upper()]
                cluster = config.build_adv_type(metric_cluster, args)
                # Set default metric for clustering
                metric = args.flitsr_default_metric
            if (cluster is not None):
                spectrums = cluster.cluster(args.input, spectrum, args.method)
            else:
                spectrums = [spectrum]
            rankings = Rankings(spectrum.get_faults(),
                                spectrum.elements())
            # Run each sub-spectrum
            for subspectrum in spectrums:
                # Run techniques
                ranker = config.ranker(args)
                if (ranker is None and hasattr(RankerType, metric.upper())):
                    metric_ranker = RankerType[metric.upper()]
                    ranker = config.build_adv_type(metric_ranker, args)
                    metric = args.flitsr_default_metric
                elif (ranker is None):
                    ranker = config.build_adv_type(RankerType['SBFL'], args)
                ranking = ranker.rank(subspectrum, metric)
                # Compute cut-off
                if (args.cutoff_strategy):
                    ranking = compute_cutoff(args.cutoff_strategy, ranking,
                                             subspectrum, metric,
                                             args.cutoff_eval)
                rankings.append(ranking)
            # Compute and print output
            output(rankings, args.weff, args.top1, args.perc_at_n,
                   args.prec_rec, args.faults, args.collapse, csv=args.csv,
                   decimals=args.decimals, file=output_file)
            spectrum.reset()


if __name__ == "__main__":
    main()



# PYTHON_ARGCOMPLETE_OK
import math
import sys
import ast
import numpy as np
from argparse import ArgumentParser
import argcomplete
from enum import Enum
from typing import Dict, List, Tuple, Optional, Collection
from flitsr.spectrum import Spectrum
from flitsr.tie import Ties
from flitsr.suspicious import Suspicious
from flitsr import advanced


def ci_in(str1: str, col: Collection[str]):
    """ Case insensitive check for strings in a list """
    return str1.casefold() in map(str.casefold, col)


def getBumps(ties: Ties, worst_effort=False, collapse=False) -> List[float]:
    if (len(ties.faults) == 0):
        return [0.0]
    tie_iter = iter(ties)
    total = 0
    size = 0
    size = ties.size(collapse)
    bumps = [float(size)]
    try:
        while (True):
            tie = next(tie_iter)
            for f in range(1, tie.num_faults()+1):
                expect_value = tie.expected_value(f, False, collapse)
                bumps.append(total+expect_value)
            total += tie.len(collapse)
    except StopIteration:
        pass
    return bumps


def combine(results: List[Tuple[float, List[float]]]) -> List[Tuple[float, float]]:
    def err(x):
        return max(10**-6, (math.log((99*x/100) + 1)/math.log(100))*0.1)
    size = len(results)
    total = 0.0
    pointers = [0]*size
    final = [len(r[1]) for r in results]
    combined = []
    def val(results, pointer, i):
        if (pointer < len(results[i][1])):
            return results[i][1][pointer]*100/results[i][0]
    while (pointers != final):
        # get the minimum next value
        min_ = min([v for v in (val(results, pointers[i], i) for i in range(size))
                        if v is not None])
        indexes = []
        for i in range(size):
            j = 0
            while ((v := val(results, pointers[i]+j, i)) is not None):
                if (abs(min_ - v) < err(min_)):
                    j += 1
                else:
                    break
            if (j > 0):
                indexes.append((i, j))
        #curr = [(math.inf if (pointers[i] == len(results[i])) else results[i][pointers[i]])
                #for i in range(size)]
        #min_ = min(curr)
        #indexes = [i for i,x in enumerate(curr) if x - min_ < 10e-4]
        for (i, j) in indexes:
            pointers[i] += j
            total += j * (100/(len(results[i][1])*size))
        combined.append((min_, round(total, 8)))
    return combined


plot_type = Enum("plot_type", "metric mode")


def read_comb_file(comb_file: str) -> Tuple[List[str], List[str],
        Dict[Tuple[str, str], List[Tuple[float, float]]]]:
    lines = open(comb_file).readlines()
    metrics = []
    modes = []
    metric = ""
    mode = ""
    points = {}
    for line in lines:
        if (line.startswith("\t\t")):
            comb = ast.literal_eval(line.strip().split(": ")[1])
            comb.append((100.0, 100.0))
            points[(metric, mode)] = comb
        elif (line.startswith("\t")):
            mode = line.strip()
            if (mode not in modes):
                modes.append(mode)
        else:
            metric = line.strip()
            if (metric not in metrics):
                metrics.append(metric)
    return metrics, modes, points


def plot(plot_file: str, log=True, all=False, type=plot_type.metric,
         incl_metrics=None, flitsrs=None, incl_advs=None,
         save=None):
    from matplotlib import pyplot as plt
    comb_points: Dict[Tuple[str, str], List[Tuple[float, float]]] = {}
    metrics, modes, comb_points = read_comb_file(plot_file)
    if (incl_metrics is not None):
        metrics = [m for m in metrics if ci_in(m, incl_metrics)]
    if (incl_advs is not None):
        modes = [m for m in modes if ci_in(m, incl_advs)]
    points = {}
    for item in comb_points.items():
        key = item[0]
        x = [0.0]
        y = [0.0]
        for inc in item[1]:
            x.append(inc[0])
            y.append(inc[1])
        points[key] = (x, y)
    if (all):
        num_plots = 1
    elif (type == plot_type.metric):
        num_plots = len(metrics)
    else:
        num_plots = len(modes)
    plt.rcParams.update({'font.size': 10})
    fig, axs = plt.subplots(1, num_plots,
                            gridspec_kw={"left": 0.055,
                                         "bottom": 0.06,
                                         "right": 0.99,
                                         "top": 0.99,
                                         "wspace": 0.2,
                                         "hspace": 0.2} if all else {})
    if (all):
        plot_all(axs, points, metrics, modes, flitsrs, log)
    else:
        plot_sep(axs, points, type, metrics, modes, log)
    if (save is None):
        plt.show()
    else:
        plt.savefig(fname=save)
    plt.close()


def plot_all(axs, points, metrics, modes, flitsrs, log):
    import matplotlib.cm as cm
    color = list(cm.rainbow(np.linspace(0, 1, len(metrics))))
    style = [(), (6, 3), (1, 3), (1, 3, 6, 3), (3, 3), (6, 3, 3, 3)]
    marker = ['D', 'o', '^', '8', 's', 'p', '*', 'x', '+', 'v', '<', '>', 'P',
              'h', 'X', 'H', 'd', '|', '_']
    lines = []
    for (i, me) in enumerate(metrics):
        for (j, mo) in enumerate(modes):
            if ((me, mo) in points and (mo == "Base" or flitsrs is None
                                        or ci_in(me, flitsrs))):
                point = points[(me, mo)]
                # print(point)
                ls = axs.plot(point[0], point[1], dashes=style[j],
                              color=color[i], marker=marker[i], markevery=0.1)
                if (mo == "Base"):
                    lines.append(ls[0])
    # lines = axs.get_lines()[::len(modes)]
    dummy_lines = []
    for j in range(len(modes)):
        dummy_lines.append(axs.plot([], [], c="black",
                                    dashes=style[j])[0])
    t_dummy, = axs.plot([], [], marker='None', ls='None')
    labels = [r'$\mathbf{Metrics\!:}$'] + metrics + \
             [r'$\mathbf{Types\!:}$'] + modes
    all_lines = [t_dummy] + lines + [t_dummy] + dummy_lines
    axs.legend(all_lines, labels)
    # axs.set_title("")
    plot_log(axs, log)
    # plt.ylim(0, 100)
    # plt.xlim(0, 100)
    axs.grid()


def plot_sep(axs, points, type, metrics, modes, log):
    if (type == plot_type.metric):
        split, merged = metrics, modes
    else:
        split, merged = modes, metrics
    for (i, s) in enumerate(split):
        for (j, m) in enumerate(merged):
            if (type == plot_type.mode and (m, s) in points):
                point = points[(m, s)]
            elif ((s, m) in points):
                point = points[(s, m)]
            # print(point)
            axs[i].plot(point[0], point[1])
    # plt.step(x,y)
    for i in range(len(axs)):
        ax = axs[i]
        ax.legend(merged)
        ax.set_title(split[i])
        plot_log(ax, log)
        # plt.ylim(0, 100)
        # plt.xlim(0, 100)
        ax.grid()


def plot_log(ax, log):
    if (log):
        ax.set_xscale("log")
        ax.set_xticks([0.01, 0.1, 1, 10, 100], [0.01, 0.1, 1, 10, 100])


def auc_calc(points: List[Tuple[float, float]],
             cut_off=101.0) -> float:
    points = sorted(points, key=lambda x: x[0])  # sort points, sanity check
    auc = 0.0
    if (len(points) == 0 or points[-1][0] != 100.0):
        points.append((100.0, 100.0))
    for i in range(1, len(points)):
        if (points[i][0] >= cut_off):
            break
        auc += (points[i][0] - points[i-1][0]) * points[i-1][1]
    return auc


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(prog="percent_at_n")
    subparsers = parser.add_subparsers(title='Modes', description='The '
                                       'following modes are available '
                                       'for the percent_at_n command:',
                                       required=True, dest='mode')
    combine_parser = subparsers.add_parser('combine')
    combine_parser.add_argument('input_file')

    met_names = (Suspicious.getNames(True) +
                 [m.lower() for m in advanced.all_types.keys()])
    plot_parser = subparsers.add_parser('plot', help='Plots the percentage at '
                                        'n graph for the given input file '
                                        'produced by the merge script')
    plot_parser.add_argument('plot_file', help='The input file generated by '
                             'the FLITSR merge script')
    plot_parser.add_argument('-a', '--all', action='store_true',
                             help='Plots one graph containing all metrics and '
                             'advanced types. You may optionally filter the '
                             'metrics that are plotted and for which metrics '
                             'advanced types are plotted for by the --metrics '
                             'and --types options')
    plot_parser.add_argument('-m', '--metrics', nargs='+', choices=met_names,
                             help='Specifies the metrics to plot for the --all '
                             'plotting style. Allowed values are: '
                             f'[{", ".join(met_names)}]', metavar='METRIC')
    plot_parser.add_argument('-A', '--advanced-types', nargs='+',
                             metavar='TYPE', help='Specify the list of '
                             'advanced types to include when merging. By '
                             'default all available advanced types that '
                             'appear in filenames of found files are included',
                             action='extend')
    plot_parser.add_argument('-t', '--types', nargs='+', choices=met_names,
                             help='Specifies the metrics for which to plot '
                             'advanced types for using the --all plotting '
                             'style. See --metrics for allowed values',
                             metavar='METRIC')
    plot_parser.add_argument('-s', '--split-type', choices=['metric', 'type'],
                             help='Determines whether to plot separate graphs '
                             'for each metric, or for each type',
                             default='metric')
    plot_parser.add_argument('-l', '--log', action='store_true',
                             help='By default graphs are plot with both axes '
                             'in linear scale. This option enables plotting '
                             'the x-axis in log scale instead')
    plot_parser.add_argument('-S', '--save-to-file', action='store',
                             metavar='FILE', help='By default, the plotted '
                             'figure is displayed in an interactive window. '
                             'Using this option, you may instead save the '
                             'figure to a file given by FILE. The extension '
                             'of the filename is used to determine the output '
                             'format.', default=None)

    auc_parser = subparsers.add_parser('auc', help='Calculates the Area Under '
                                       'Curve (AUC) for each metric and '
                                       'advanced type in the given input file '
                                       'generated by the merge script')
    auc_parser.add_argument('comb_file', help='The input file generated by '
                            'the FLITSR merge script')
    auc_parser.add_argument('-m', '--metrics', nargs='+', choices=met_names,
                            help='Specifies the subset of metrics to '
                            'calculate AUC values for. Allowed values are: '
                            f'[{", ".join(met_names)}]', metavar='METRIC')
    auc_parser.add_argument('-c', '--cutoff', action='store', type=float,
                            help='Specifies the percentage cut-off point to '
                            'use for the AUC calculations', default='101.0')
    argcomplete.autocomplete(parser)
    return parser


def main(argv: Optional[List[str]] = None):
    parser = get_parser()
    args = parser.parse_args(argv)
    if (args.mode == "combine"):
        infile = args.input_file
        lines = open(infile).readlines()
    elif (args.mode == "plot"):
        plot(args.plot_file, log=args.log, all=args.all, type=args.split_type,
             incl_metrics=args.metrics, flitsrs=args.types,
             incl_advs=args.advanced_types, save=args.save_to_file)
    elif (args.mode == "auc"):
        comb_file = args.comb_file
        metrics, modes, points = read_comb_file(comb_file)
        cutoff = args.cutoff
        if (args.metrics):
            metrics = args.metrics
        for metric in metrics:
            base = auc_calc(points[(metric, "Base metric")], cutoff)
            flitsr = auc_calc(points[(metric, "FLITSR")], cutoff)
            flitsr_m = auc_calc(points[(metric, "FLITSR*")], cutoff)
            print("{} FLITSR imprv: {:.6%}".format(metric, flitsr/base))
            print("{} FLITSR* imprv: {:.6%}".format(metric, flitsr_m/base))


if __name__ == "__main__":
    main()



from fractions import Fraction
from typing import Any, Dict, Tuple, Set
from math import comb, factorial
from flitsr.spectrum import Spectrum
from flitsr.tie import Ties


def precision(n: Any, ties: Ties, perc=False, worst_effort=False,
              collapse=False) -> float:
    if (len(ties.faults) == 0):
        return 0.0
    fault_num, total = method(n, ties, perc, worst_effort, collapse)
    return fault_num/total


def recall(n: Any, ties: Ties, perc=False, worst_effort=False,
           collapse=False) -> float:
    if (len(ties.faults) == 0):
        return 0.0
    fault_num, total = method(n, ties, perc, worst_effort, collapse)
    return fault_num/len(ties.faults)


def method(n: Any, ties: Ties, perc: bool = False, worst_effort: bool = False,
           collapse: bool = False) -> Tuple[float, int]:
    size = ties.size(collapse)
    if (n == "b"):
        n = -1
    elif (n == "f"):
        n = len(ties.faults)
    elif (perc):
        n = n * size
    tie_iter = iter(ties)
    total = 0
    fault_num = 0.0
    try:
        while (total < n):
            tie = next(tie_iter)
            faults = tie.active_faults(collapse)
            add = 0.0
            if (total+tie.len(collapse) > n and tie.num_faults() > 0):
                p = int(n - total)  # num of statements to inspect in the tie
                Ntot = tie.len(collapse)  # total num of statements in the tie
                expval = 0.0
                # sum the expected values of finding each fault
                for f in faults:
                    Ni = len(faults[f])
                    # Expected value of each fault is {1, 0} <=> probability
                    # P(selecting one or more locations of this fault)
                    # <=> 1 - P(selecting no locations of this fault)
                    expval += 1 - Fraction(comb(Ntot-Ni, p), comb(Ntot, p))
                add += float(expval)  # add the expected # of faults found in p
                total += p
            else:
                add = tie.num_faults()
                total += tie.len(collapse)
            fault_num += add
    except StopIteration:
        pass
    return fault_num, total



import random
from typing import List, Set, Optional, Dict, Iterator, Iterable, Any
from enum import Enum, auto
from flitsr.spectrum import Spectrum
from flitsr.errors import warning
import copy


class Tiebrk(Enum):
    """ An `Enum` for the tie-breaking method to be used. """

    EXEC = auto()
    """ Tie break only using execution counts. """

    RNDM = auto()
    """ Tie break randomly. Note, this will cause non-deterministic output. """

    ORIG = auto()
    """ Tie break using the original SBFL ranking, then execution counts. """


class Rank:
    """
    An `Entity <flitsr.spectrum.Spectrum.Entity>` from a spectrum, together
    with its score and execution count.
    """
    def __init__(self, entity: Spectrum.Entity, score: float,
                 exec_count: int):
        self.entity = entity
        self.score = score
        self.exec = exec_count

    def __str__(self):
        return "(" + str(self.entity) + ", " + str(self.score) + ")"

    def __repr__(self):
        return str(self)


class Ranking:
    """
    A singular ranking, containing ranked `Spectrum.Entity
    <flitsr.spectrum.Spectrum.Entity>`.
    """

    def __init__(self, tiebrk: Tiebrk = Tiebrk.ORIG):
        self._ranks: List[Rank] = []
        self._tiebrk = tiebrk
        self.entity_map: Dict[Spectrum.Entity, Rank] = {}
        self.place = 0

    def __getitem__(self, index: int):
        return self._ranks[index]

    def get_rank(self, entity: Spectrum.Entity, check_sub_group=False) -> Rank:
        """
        Return the `Rank` of the given `Spectrum.Entity
        <flitsr.spectrum.Spectrum.Entity>`. If the `check_sub_group`
        option is given, try to find if the given entity is contained within
        another group in the ranking (either as a sub-group or as an element).

        Args:
          entity: Spectrum.Entity: The entity to return the rank for
          check_sub_group:  (Default value = False) Whether to check for the
          `entity` as a subgroup of entities in the ranking.

        Returns:
          The `Rank` of the given entity.

        """
        try:
            return self.entity_map[entity]
        except KeyError as keyerror:
            if (check_sub_group):
                # Before exiting, first check if we can find a super-group
                is_group = isinstance(entity, Spectrum.Group)
                for key in self.entity_map:
                    # If a group, check if it is a sub-group
                    if (is_group):
                        if (isinstance(key, Spectrum.Group) and
                            key.is_subgroup(entity)):  # type:ignore
                            return self.entity_map[key]
                    # If an element, check if in any group
                    elif (entity in key):
                        return self.entity_map[key]
            # if no super-group can be found, raise the KeyError
            raise keyerror

    def sort(self, reverse: bool):
        """
        Re-sort this `Ranking` in-place by their scores, using the `Tiebrk`
        method set.

        Args:
          reverse: bool: Whether to sort by reverse order of scores.
        """
        if (self._tiebrk == Tiebrk.EXEC):  # Sorted by execution counts
            self._ranks.sort(key=lambda x: x.exec, reverse=reverse)
        elif (self._tiebrk == Tiebrk.RNDM):  # random ordering
            random.shuffle(self._ranks)
        elif (self._tiebrk == Tiebrk.ORIG):  # original ranking tie break
            if (orig is not None):  # sort by original rank then exec count
                try:
                    self._ranks.sort(key=lambda x:
                                     orig.get_rank(x.entity, True).exec,
                                     reverse=reverse)
                    self._ranks.sort(key=lambda x:
                                     orig.get_rank(x.entity, True).score,
                                     reverse=reverse)
                except KeyError:
                    # if sorting by orig fails, print a warning and continue
                    warning("Could not sort by original ranking, despite it "
                            "being set")
                    pass
            else:  # if no orig, still sort by current execution count
                self._ranks.sort(key=lambda x: x.exec, reverse=reverse)
        self._ranks.sort(key=lambda x: x.score, reverse=reverse)

    def append(self, entity: Spectrum.Entity, score: float, exec_count: int):
        """
        Append the given `entity` to the ranking, with the given `score` and
        execution count `exec_count`.

        Args:
          entity: Spectrum.Entity: The entity to append to this `Ranking`.
          score: float: The score of the given `entity`.
          exec_count: int: The execution count of the given `entity`.
        """
        created = Rank(entity, score, exec_count)
        self._ranks.append(created)
        self.entity_map[entity] = created

    def extend(self, ranks: List[Rank]):
        """
        Extend this `Ranking` by adding all `Rank` in the `List` `ranks`.

        Args:
          ranks: List[Rank]: The `List` of ranks to add to this `Ranking`.
        """
        self._ranks.extend(ranks)
        for rank in ranks:
            self.entity_map[rank.entity] = rank

    def has_entity(self, entity: Spectrum.Entity) -> bool:
        """
        Return whether this `Ranking` contains the given `entity`. Does not
        check if the entity is a subgroup.

        Args:
          entity: Spectrum.Entity: The entity to check for.

        Returns:
          True if `entity` is directly in this `Ranking`, False otherwise.
        """
        return entity in self.entity_map

    def __iter__(self):
        return iter(self._ranks)

    def __len__(self):
        return len(self._ranks)

class Rankings:
    """
    A collection of `Ranking` objects, which share a non-overlapping set of
    elements from a `Spectrum <flitsr.spectrum.Spectrum>`.
    """
    def __init__(self, faults: Dict[Any, Set[Spectrum.Element]],
                 elements: List[Spectrum.Element],
                 rankings: Optional[Iterable[Ranking]] = None):
        self._faults = faults
        self._all_elems = elements
        self._rankings: List[Ranking] = []
        if (rankings is not None):
            self._rankings.extend(rankings)

    def faults(self) -> Dict[Any, Set[Spectrum.Element]]:
        """
        Return a dictionary of all faults in these rankings. See
        `Spectrum.get_faults <flitsr.spectrum.Spectrum.get_faults>` for a
        description of the return value.
        """
        return copy.deepcopy(self._faults)

    def elements(self) -> List[Spectrum.Element]:
        """ Return the global list of elements for all rankings. """
        return self._all_elems

    def rankings(self) -> List[Ranking]:
        """ Return a list of all the `Ranking`. """
        return self._rankings

    def append(self, ranking: Ranking):
        """
        Append the given `ranking` to this collection.

        Args:
          ranking: Ranking: The ranking to add.
        """
        self._rankings.append(ranking)

    def __iter__(self) -> Iterator:
        return iter(self._rankings)

    def __len__(self) -> int:
        return len(self._rankings)


orig: Optional[Ranking] = None


def set_orig(ranking: Ranking):
    """
    Set the original SBFL `Ranking` to be used in tie-breaking (see `Tiebrk`).
    Args:
      ranking: Ranking:

    Returns:

    """
    global orig
    orig = copy.deepcopy(ranking)


def unset_orig():
    """ """
    global orig
    orig = None



import re
from typing import Tuple, List, Set, Union, Any, Dict
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking, Rankings


def read_any_ranking(ranking_file: str, method_level=False) -> Rankings:
    """
    Guess the ranking from the contents of the `ranking_file` and read it in.

    Args:
      ranking_file: str: The ranking input file to read in.
      method_level:  (Default value = False) Whether the ranking file is method
        level.

    Returns:
      A `Rankings <flitsr.ranking.Rankings>` object containing the single
      read-in ranking.
    """
    f = open(ranking_file)
    if (f.readline().startswith("Faulty grouping")):
        return read_flitsr_ranking(ranking_file)
    else:
        return read_gzoltar_ranking(ranking_file, method_level)


def read_gzoltar_ranking(ranking_file: str, method_level=False) -> Rankings:
    """
    Read in a GZoltar formatted ranking.
    """
    f = open(ranking_file)
    ranking = Ranking()
    bugs = 0
    methods: Dict[Tuple[str, str], Spectrum.Element] = {}
    all_faults: Dict[Any, Set[Spectrum.Element]] = {}
    elements: List[Spectrum.Element] = []
    f.readline()
    for i, line in enumerate(f):
        line = line.strip()
        score = float(line[line.index(";")+1:])
        name = line[:line.index(";")]
        l = name.strip().split(':')
        r = re.search("(.*)\\$(.*)#([^:]*)", l[0])
        if (r is None):
            raise ValueError("Incorrectly formatted line \"" + line +
                             "\" when reading input ranking file")
        faults = []
        if (len(l) > 2):
            if (not l[2].isdigit()):
                faults = [bugs]
            else:
                faults = []
                for b in l[2:]:
                    faults.append(int(b))
            bugs += 1
        details = [r.group(1)+"."+r.group(2), r.group(3), l[1]]
        # Create or fetch the element
        if (not method_level or (details[0], details[1]) not in methods):
            elem = Spectrum.Element(details, len(elements), faults)
            elements.append(elem)
            if (method_level):
                methods[(details[0], details[1])] = elem
        else:
            elem = methods[(details[0], details[1])]
            for fault in faults:
                if (fault not in elem.faults):
                    elem.faults.append(fault)
        # Add/Update the method's score
        if (ranking.has_entity(elem)):
            rank_elem = ranking.get_rank(elem)
            rank_elem.score = max(rank_elem.score, score)
        else:
            ranking.append(elem, score, 0)
        # Update faults
        if (elem.faults):
            for fault in elem.faults:
                all_faults.setdefault(fault, set()).add(elem)
    return Rankings(all_faults, elements, [ranking])


def read_flitsr_ranking(ranking_file: str) -> Rankings:
    """
    Read in a ``flitsr`` formatted ranking.
    """
    f = open(ranking_file)
    ranking = Ranking()
    all_faults: Dict[Any, Set[Spectrum.Element]] = {}
    elements: List[Spectrum.Element] = []
    num_locs = 0  # number of reported locations (methods/lines)
    i = 0  # number of actual lines
    line = f.readline()
    while (line != ""):
        line = line.strip()
        score: Union[int, float]
        str_score = line[line.index(": ")+2:line.index(" [")]
        if (str_score.isdigit()):
            score = int(str_score)
        else:
            score = float(str_score)
        line = f.readline().strip()
        group_elems = []
        while (not line.startswith("]")):
            # read in ranked element (old or new FLITSR format)
            m = re.fullmatch("\\s*(?:\\([0-9]+\\)\\s*)?(\\S*)\\s*(?:\\(FAULT ([0-9.,]+)\\))?", line)
            if (m is None):
                # if normal format fails, try DUA format
                m = re.fullmatch("\\s*(?:\\([0-9]+\\)\\s*)?(\\S*\\s\\S*\\s\\S*)\\s*(?:\\(FAULT ([0-9.,]+)\\))?", line)
                if (m is None):
                    raise ValueError("Incorrectly formatted line \"" + line +
                                     "\" when reading input ranking file")
            details = m.group(1).split('|')
            if (m.group(2)):
                faults = [int(i) if i.isdecimal() else float(i)
                          for i in m.group(2).split(',')]
            else:
                faults = []
            elem = Spectrum.Element(details, len(elements), faults)
            elements.append(elem)
            for fault in faults:
                all_faults.setdefault(fault, set()).add(elem)
            group_elems.append(elem)
            i += 1
            line = f.readline().strip()
        group = Spectrum.Group(group_elems)
        ranking.append(group, score, 0)
        num_locs += 1
        line = f.readline().strip()
    rankings = Rankings(all_faults, elements, [ranking])
    return rankings



from __future__ import annotations
import pkgutil
from flitsr import advanced
from flitsr.advanced.refiner import Refiner
from flitsr.advanced.cluster import Cluster
from flitsr.advanced.ranker import Ranker
from enum import Enum
import importlib
from typing import Union, Optional, Any, Dict, Type, TYPE_CHECKING, overload
import sys
if TYPE_CHECKING:
    from flitsr.args import Args
from deprecated.sphinx import deprecated, versionadded, versionchanged

_rankers = {}
_clusters = {}
_refiners = {}
all_types = {}
all_types_print = []


def register_ranker(cls):
    if (cls.__name__.upper() in all_types):
        raise ValueError(f'Duplicate advanced types for name "{cls.__name__}"')
    _rankers[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


def register_cluster(cls):
    if (cls.__name__.upper() in all_types):
        raise ValueError(f'Duplicate advanced types for name "{cls.__name__}"')
    _clusters[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


def register_refiner(cls):
    if (cls.__name__.upper() in all_types):
        raise ValueError(f'Duplicate advanced types for name "{cls.__name__}"')
    _refiners[cls.__name__.upper()] = cls
    all_types[cls.__name__.upper()] = cls


# load local advanced types
__all__ = [m[1] for m in pkgutil.iter_modules(advanced.__path__)]
for module in __all__:
    importlib.import_module('.'+module, package=__name__)
# load plugin advanced types
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points
adv_entry_points = entry_points(group='flitsr.advanced')
for adv_ep in adv_entry_points:
    adv_ep.load()

RefinerType = Enum('RefinerType', _refiners)  # type:ignore
# module=advanced, qualname='advanced.RefinerType')
ClusterType = Enum('ClusterType', _clusters)  # type:ignore
# module=advanced, qualname='advanced.ClusterType')
RankerType = Enum('RankerType', _rankers)  # type:ignore
# module=advanced, qualname='advanced.RankerType')

AdvType = Union[RefinerType, ClusterType, RankerType]

for cls in all_types.values():
    if (hasattr(cls, '__print_name__')):
        all_types_print.append(cls.__print_name__.lower())
    else:
        all_types_print.append(cls.__name__.lower())


@versionchanged(version='2.3.0', reason='Replaced the `ranker`, `cluster`, '
                'and `refiner` attributes with functions constructing them')
class Config:
    """
    The advanced type configuration for a particular run of SBFL methods.

    Optionally contains a single Refiner, Cluster, and/or Ranker, as well as
    any arguments related to these advanced types. Multiple of these Config
    objects can be created in the same execution.
    """
    @versionchanged(version='2.4.0',
                    reason='Added the `args` optional argument')
    def __init__(self, ranker: Optional[RankerType] = None,
                 cluster: Optional[ClusterType] = None,
                 refiner: Optional[RefinerType] = None,
                 args: Optional[Dict[str, Dict[str, Any]]] = None):
        """
        Construct a `Config` object with the given ranker, cluster, and/or
        refiner (all optional).

        Args:
          ranker:  The (optional) `RankerType` to include in this config
          cluster: The (optional) `ClusterType` to include in this config
          refiner: The (optional) `RefinerType` to include in this config
          args:    If given, the dictionary containing additional arguments for
            each of the advanced types. The dictionary is indexed by the name
            of the advanced type, with the value being another dictionary with
            argument names and values. For example: `{'MULTI': {'cutoff': 5}}`
            will set the `cutoff <flitsr.advanced.flitsr.Multi.__init__>`
            parameter of the `MULTI <flitsr.advanced.flitsr.Multi>` RankerType.
        """
        self._cluster = cluster
        self._ranker = ranker
        self._refiner = refiner
        if (args is None):
            self._args = {}
        else:
            self._args = args

    def __str__(self, params: Optional[str] = ""):
        return self.get_str(filename=True, params=params)

    @versionadded(version='2.4.0')
    def get_str(self, filename=False, params: Optional[str] = None):
        """
        Returns a string description of the advanced types of this config
        object that can be used either in the name of the output file
        (`filename=True`), or for other script use (`filename=False`).

        Args:
          params: (Default value = None) If and how to include
            advanced type parameters in the output string. If None,
            parameters will not be included in the output string. If the
            empty string ("") is given, parameters will be included in the
            format "{k}-{v}" (for filenames) or "{k}={v}" (for
            non-filenames), where "{k}" is the parameter name, and "{v}" the
            value. Alternatively, a different format can be specified by
            giving a string using the same "{k}", "{v}" format.

        Returns:
          A string description of this `Config`.
        """
        c = '_' if filename else '+'
        components = [self._ranker or RankerType['SBFL'], self._refiner,
                      self._cluster]
        string = c.join(self._get_adv_name(c, filename, params)
                        for c in components if c is not None)
        return string

    def _get_args(self, adv_type: AdvType):
        return self._args.get(adv_type.name.upper(), {})

    def _get_params(self, adv_type: AdvType, args: Args):
        arg_params = args.get_arg_group(adv_type.name)
        custom_params = self._get_args(adv_type)
        refiner_params = {**arg_params, **custom_params}
        return refiner_params

    @versionadded(version='2.4.0')
    def set_arg(self, adv_type: AdvType, arg_name: str, arg_value: Any):
        """
        Set the argument given by `arg_name` with `arg_value` for the given
        advanced type.

        Args:
          adv_type: The advanced type to set the argument for.
          arg_name: str: The name of the argument to set
          arg_value: `Any <typing.Any>`: The value of the argument to set
        """
        self._args.setdefault(adv_type.name, {})[arg_name] = arg_value

    @versionadded(version='2.4.0')
    def refiner(self, args: Args) -> Optional[Refiner]:
        """
        Build and return this `Config`'s `RefinerType`, if available.

        Args:
          args: The global flitsr arguments to use when constructing this
            `Config`'s `RefinerType`.

        Returns:
          This `Config`'s constructed `Refiner <flitsr.advanced.refiner.Refiner>`
          object, or `None` if no `RefinerType` has been set.
        """
        return self.build_adv_type(self._refiner, args)

    @versionadded(version='2.4.0')
    def cluster(self, args: Args) -> Optional[Cluster]:
        """
        Build and return this `Config`'s `ClusterType`, if available.

        Args:
          args: The global flitsr arguments to use when constructing this
            `Config`'s `ClusterType`.

        Returns:
          This `Config`'s constructed `Cluster <flitsr.advanced.cluster.Cluster>`
          object, or `None` if no `ClusterType` has been set.
        """
        return self.build_adv_type(self._cluster, args)

    @versionadded(version='2.4.0')
    def ranker(self, args: Args) -> Optional[Ranker]:
        """
        Build and return this `Config`'s `RankerType`, if available.

        Args:
          args: The global flitsr arguments to use when constructing this
            `Config`'s `RankerType`.

        Returns:
          This `Config`'s constructed `Ranker <flitsr.advanced.ranker.Ranker>`
          object, or `None` if no `RankerType` has been set.
        """
        return self.build_adv_type(self._ranker, args)

    @overload
    def build_adv_type(self, adv_type: RankerType, args: Args) -> Ranker: ...

    @overload
    def build_adv_type(self, adv_type: ClusterType, args: Args) -> Cluster: ...

    @overload
    def build_adv_type(self, adv_type: RefinerType, args: Args) -> Refiner: ...

    @overload
    def build_adv_type(self, adv_type: None, args: Args) -> None: ...

    @versionadded(version='2.4.0')
    def build_adv_type(self, adv_type: Optional[AdvType], args: Args) -> Union[
            Refiner, Cluster, Ranker, None]:
        """
        Constructs the advanced type given by `adv_type` using the parameters
        in both the given `args` and this `Config`'s arguments.

        Args:
          adv_type: The advanced type to construct, which is one of
            `RefinerType`, `ClusterType`, or `RankerType`.
          args: The global flitsr arguments to use when constructing the
            given advanced type.

        Returns:
          The constructed advanced type object, which corresponds to the given
          `adv_type` argument.
        """
        if (adv_type is None):
            return None
        params = self._get_params(adv_type, args)
        mthd = adv_type.value(**params)
        return mthd

    @versionadded(version='2.4.0')
    def get_concrete(self, adv_type: Union[Type[RefinerType],
                     Type[ClusterType], Type[RankerType]]) -> Optional[str]:
        """
        Returns a string name of this `Config`'s concrete advanced type for the
        given category (Refiner, Cluster or Ranker), if it is available.
        For example, if RankerType was given, and this Config contained the
        SBFL RankerType, then "SBFL" would be returned. If there is no concrete
        advanced type for the given category, None is returned.

        Args:
          adv_type: The advanced type category to retrieve.

        Returns:
          The name of the concrete implementation for the given advanced type
          in this `Config` object, or None if no concrete implementation is
          available.
        """
        typ: Optional[AdvType] = None
        if (adv_type is RefinerType):
            typ = self._refiner
        elif (adv_type is ClusterType):
            typ = self._cluster
        elif (adv_type is RankerType):
            typ = self._ranker
        if (typ is None):
            return None
        else:
            return typ.name

    def _get_adv_name(self, typ: AdvType, filename: bool,
                      params: Optional[str] = None) -> str:
        # first get args (if needed and available)
        args = self._get_args(typ)
        # Always add params for non-filename
        if (not filename and params is None):
            params = ""
        if (params is not None and args is not None and len(args) > 0):
            if (params == ""):
                if (filename):
                    params = "{k}-{v}"
                else:
                    params = "{k}={v}"
            param_set = []
            for k, v in args.items():
                if (filename):
                    k = k.replace("_", "-")
                param_set.append(params.replace("{k}", k)
                                       .replace("{v}", v))
            if (filename):
                a = '_' + '_'.join(param_set)
            else:
                a = '(' + ','.join(param_set) + ')'
        else:
            a = ''
        # then get name
        if (filename and hasattr(typ.value, '__print_name__')):
            name = typ.value.__print_name__.lower()
        elif (filename):
            name = typ.name.lower()
        else:
            name = typ.name.upper()
        return name + a

    @versionadded(version='2.4.0')
    def __repr__(self, params: Optional[str] = ""):
        return self.get_str(filename=False, params=params)

    @versionchanged(version='2.4.0',
                    reason='Added the `params` optional argument')
    def get_file_name(self, params: Optional[str] = None):
        """
        Synonym for `Config.get_str(filename=True, ...) <get_str>`.

        Returns a string description of the advanced types of this config
        object that can be used in the name of the output file.

        Args:
          params: (Default value = None) Whether to include advanced type
            parameters in the file name. See the `get_str` method for a
            description of the format.

        Returns:
          A string description of this `Config` for filename use.
        """
        return self.get_str(filename=True, params=params)



from __future__ import annotations
from abc import ABC, abstractmethod
from flitsr.spectrum import Spectrum
from flitsr.ranking import Ranking
from flitsr import advanced


class Ranker(ABC):
    """
    A `Ranker` technique takes a `Spectrum <flitsr.spectrum.Spectrum>` and
    forms a `Ranking <flitsr.ranking.Ranking>` of the elements in the the
    spectrum.
    """

    @abstractmethod
    def rank(self, spectrum: Spectrum, base_metric: str) -> Ranking:
        """
        Provides the ranking functionality for the `Ranker`. Takes a `Spectrum
        <flitsr.spectrum.Spectrum>` and a `base_metric` and returns a `Ranking
        <flitsr.ranking.Ranking>` of the elements in the spectrum. Note that the
        technique does not need to use the `base_metric`.

        Args:
          spectrum: Spectrum: The spectrum whose elements to rank
          base_metric: str: The name of the SBFL metric to optionally use within
            the technique to rank the elements.

        Returns:
          A `Ranking <flitsr.ranking.Ranking>` of the elements in the
          `spectrum`.
        """
        pass

    def __init_subclass__(cls):
        advanced.register_ranker(cls)



import importlib
import pkgutil
import sys
from enum import Enum
from flitsr import input

_inputs = {}


def register_input(cls):
    _inputs[cls.__name__.upper()] = cls


# load local inputs
__all__ = [m[1] for m in pkgutil.iter_modules(input.__path__)]
for module in __all__:
    importlib.import_module('.'+module, package=__name__)
# load plugin inputs
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points
adv_entry_points = entry_points(group='flitsr.input')
for adv_ep in adv_entry_points:
    adv_ep.load()

InputType = Enum('InputType', _inputs,  # type:ignore
                 module=input, qualname='input.InputType')



import sys
from os import path as osp
from io import TextIOWrapper
import re
import mmap
from flitsr.spectrum import Spectrum, Outcome
from flitsr.spectrumBuilder import SpectrumBuilder, TestKeyError, ElemKeyError
from flitsr.errors import error
from flitsr.input.input_reader import Input
from flitsr.split_faults import split_spectrum_faults, NoFaultsError


class TCM(Input):
    def get_run_file_name(self, input_path: str):
        return re.sub("\\.\\w+$", ".run", input_path)

    def construct_details(self, f: TextIOWrapper, method_level: bool,
                          sb: SpectrumBuilder):
        """
        Fills the spectrum object with elements read in from the open file 'f'.
        """
        bugs = 0
        line = f.readline()
        while (not line == '\n'):
            m = re.fullmatch('((?:[^ ]| [^|])+)(?: \\| (.+))?', line.rstrip())
            if (m is None):
                error("incorrectly formatted element line in input file:",
                      line, "terminating...")
            else:
                faults = []
                if (m.group(2) is not None):
                    bs = m.group(2).split(' | ')
                    if (not bs[0].isdigit()):
                        faults = [bugs]
                    else:
                        faults = []
                        for b in bs:
                            faults.append(int(b))
                    bugs += 1
                if (method_level):
                    details_m = re.match('([^:]+):([^:]+):([0-9]+)',
                                         m.group(1))
                    if (details_m is None):
                        error("Incorrectly formatted line \"" + line +
                              "\" for method level evaluation")
                    else:
                        details = list(details_m.groups())
                else:
                    details = m.group(1).split(":")
                sb.addElement(details, faults)
                line = f.readline()

    def construct_tests(self, f: TextIOWrapper, sb: SpectrumBuilder):
        line = f.readline()
        while (not line == '\n'):
            m = re.fullmatch('([^ ]+) (PASSED|FAILED|ERROR)( .*)?',
                             line.rstrip())
            if (m is None):
                error("incorrectly formatted test line in input file:", line,
                      "terminating...")
            else:
                sb.addTest(m.group(1), Outcome[m.group(2)])
            line = f.readline()

    def fill_spectrum(self, f: TextIOWrapper, sb: SpectrumBuilder):
        for t, test in enumerate(sb.get_tests()):
            line = f.readline()
            if (line == ''):
                error('Incorrect number of matrix lines', f'({t})',
                      'in input file, terminating...')
            arr = line.rstrip().split()
            for i in range(0, len(arr), 2):
                try:
                    sb.addExecution(t, int(arr[i]))
                except ElemKeyError:
                    error(f'Incorrect element reference "{arr[i]}"',
                          f'at column {i*2+1} on line {t} of the matrix',
                          'component in the input file, terminating...')
                except TestKeyError:
                    error('Incorrect number of matrix lines', f'({t})',
                          'in input file, terminating...')

    def read_spectrum(self, input_path: str, split_faults: bool,
                      method_level=False) -> Spectrum:
        sb = SpectrumBuilder(method_level)
        file = open(input_path)
        exec_checks = {'#tests': False, '#uuts': False, '#matrix': False}
        while (True):
            line = file.readline()
            if (line == '' or not line.startswith("#")):
                break
            elif (line.startswith("#metadata")):
                while (not line == '\n'):
                    line = file.readline()
            elif (line.startswith("#tests")):
                if (exec_checks['#tests'] is True):
                    print("WARNING: duplicate #tests component in input file",
                          file=sys.stderr)
                # Constructing the spectrum
                self.construct_tests(file, sb)
                exec_checks['#tests'] = True
            elif (line.startswith("#uuts")):
                if (exec_checks['#uuts'] is True):
                    print("WARNING: duplicate #uuts component in input file",
                          file=sys.stderr)
                # Getting the details of the project
                self.construct_details(file, method_level, sb)
                exec_checks['#uuts'] = True
            elif (line.startswith("#matrix")):
                if (exec_checks['#matrix'] is True):
                    print("WARNING: duplicate #matrix component in input file",
                          file=sys.stderr)
                # Filling the spectrum
                self.fill_spectrum(file, sb)
                exec_checks['#matrix'] = True
            else:
                print("ERROR: Incorrectly formatted input file at line:", line,
                      "terminating...", file=sys.stderr)
                quit()
        file.close()
        if (sum(exec_checks.values()) != 3):
            missing = [e[0] for e in exec_checks.items() if e[1] is False]
            error(f"Input file missing components: {missing}, terminating...")
        spectrum = sb.get_spectrum()
        del sb
        # Split fault groups if necessary
        if (split_faults):
            try:
                split_spectrum_faults(spectrum)
            except NoFaultsError:
                error(f"No exposable faults in {input_path}, terminating...")
        return spectrum

    @staticmethod
    def check_format(input_path: str) -> bool:
        if (osp.isfile(input_path)):
            with open(input_path, 'rb', 0) as file:
                with mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as f:
                    return f.find(b'#uuts') != -1 and f.find(b'#tests') != -1 \
                        and f.find(b'#matrix') != -1
        return False

    @classmethod
    def write_spectrum(cls, spectrum: Spectrum, output_path: str):
        """ Output the spectrum in TCM format """
        with open(output_path, 'w') as file:
            type_ = cls.get_type()
            print("#tests", file=file)
            for test in spectrum.tests():
                print(test.name, test.outcome.name, file=file)
            print(file=file)
            print("#uuts", file=file)
            # TODO: change _elements below to elements()
            for elem in spectrum._elements:
                print(elem.output_str(type_=type_), file=file)
            print(file=file)
            print("#matrix", file=file)
            for test in spectrum.tests():
                first = True
                for elem_id, elem in enumerate(spectrum._elements):
                    if (spectrum[test][elem]):
                        print(("" if first else " ")+str(elem_id), "1", end="",
                              file=file)
                        first = False
                print(file=file)

    @classmethod
    def get_elem_separators(cls):
        return ['.', ':', ':', ' | ']


if __name__ == "__main__":
    from flitsr.output import print_spectrum
    d = sys.argv[1]
    tinput = TCM()
    spectrum = tinput.read_spectrum(d, False)
    print_spectrum(spectrum)



import subprocess
import re
import copy
import os
from typing import List
import tempfile
from importlib import resources
from flitsr.spectrum import Spectrum
from flitsr.advanced.cluster import Cluster
from flitsr.advanced.attributes import choices
from flitsr.input import InputType


class Parallel(Cluster):
    """
    Run one of the parallel debugging algorithms on the spectrum to
    produce multiple spectrums, and process all other options on
    each spectrum.
    """

    @choices('parType', ['bdm', 'msp', 'hwk', 'vwk'])
    def __init__(self, parType: str = 'msp'):
        """
        Constructs a Parallel `Cluster <flitsr.advanced.cluster.Cluster>`
        object. Takes a partition type `parType`, which is the algorithm to run.
        The choices are: 'bdm', 'msp', 'hwk', and 'vwk'.
        """
        self.parType = parType

    def cluster(self, inp_file: str, spectrum: Spectrum,
                method_lvl=False) -> List[Spectrum]:
        """
        Run one of the Parallel clustering algorithms over the given spectrum.

        Args:
          inp_file: str: The file path of the input file containing the spectrum.
          spectrum: Spectrum: The input spectrum.
          method_lvl:  (Default value = False) Whether the input spectrum is
            method level.

        Returns:
          A collection of subspectra formed by decomposing the input spectrum.
        """
        # write the spectrum to a temporary file
        tmp_fd, tmp_name = tempfile.mkstemp(text=True)
        inp_file = tmp_name
        InputType['TCM'].value.write_spectrum(spectrum, tmp_name)
        # Run the parallel algorithm
        spectrums = self.partition_table(inp_file, spectrum)
        # close and remove temporary file
        os.close(tmp_fd)
        os.remove(tmp_name)
        return spectrums

    def partition_table(self, d: str, spectrum: Spectrum) -> List[Spectrum]:
        """
        Partitions the given `spectrum` using one of the parallel algorithms
        into multiple decomposed spectra.

        Args:
          d: str: The location path for the input file containing the spectrum.
          spectrum: Spectrum: The input spectrum.

        Returns:
          A collection of spectra which are decompositions of the input
          spectrum.
        """
        jar_name = 'parallel-1.0-SNAPSHOT-jar-with-dependencies.jar'
        with resources.path('flitsr.advanced', jar_name) as jar_file:
            output = subprocess.check_output(['java', '-jar', str(jar_file), d,
                                              self.parType]).decode('utf-8')
        partitions = re.split("partition \\d+\n", output)[1:]
        spectrums: List[Spectrum] = []
        for partition in partitions:
            new_spectrum = copy.deepcopy(spectrum)
            tests = [int(i) for i in partition.strip().split("\n")]
            toRemove = set()
            all_tests = spectrum.tests()
            # keep failing + all passing
            for test in spectrum.failing():
                ind = all_tests.index(test)
                if (ind not in tests):
                    toRemove.add(test)
            for test in toRemove:
                new_spectrum.remove_test(test, bucket=None)  # hard remove test
            self.trim_groups(new_spectrum)
            spectrums.append(new_spectrum)
        return spectrums

    def trim_groups(self, spectrum):
        """
        Remove groups in the spectrum in place that are not in this block (i.e.
        ef = 0).

        Args:
          spectrum: The spectrum for which to trim the groups.
        """
        toRemove = set()
        for group in spectrum.groups():
            if (spectrum.f[group] == 0):
                toRemove.add(group)
        for group in toRemove:
            spectrum.remove_group(group)



import copy
from fractions import Fraction
from itertools import permutations, chain
from math import factorial, ceil
from typing import List, Any, Optional, Set, Dict
from flitsr.spectrum import Spectrum
from flitsr.ranking import Rankings, Ranking, Rank
from deprecated.sphinx import deprecated, versionadded


class Tie:
    def __init__(self):
        self._elems: Set[Spectrum.Element] = set()
        self._group_len = 0
        self._fault_locs: Dict[Any, Set[Spectrum.Element]] = {}
        self._fault_groups: Dict[Any, Set[Spectrum.Element]] = {}
        self._num_faults = 0
        self._num_fault_locs = 0
        self._num_fault_groups = 0
        self._all_fault_locs: Dict[Spectrum.Element, Set[Any]] = {}
        self._all_fault_groups: Dict[Spectrum.Element, Set[Any]] = {}

    def len(self, collapse=False) -> int:
        """
        Return either the number of groups (if collapsed), or number of
        elements in this tie (if not collapsed).
        """
        if (collapse):
            return self._group_len
        else:
            return len(self._elems)

    def elems(self) -> Set[Spectrum.Element]:
        """Return the set of all the elements in this tie"""
        return self._elems

    @versionadded(version='2.4.0')
    def active_faults(self, collapse=False) \
            -> Dict[Any, Set[Spectrum.Element]]:
        """
        Return a dictionary of all faults seen for the fime time in this tie,
        along with their fault locations (either groups or elements). This is a
        subset of the ``Ties.faults`` dictionary.
        """
        if (collapse):
            return self._fault_groups
        else:
            return self._fault_locs

    def num_faults(self) -> int:
        """
        Return the number of unique faults found for the first time in this
        tie.
        """
        return self._num_faults

    def num_fault_locs(self, collapse=False) -> int:
        """
        Return the total number of faulty locations (either groups or elements)
        in this tie. NOTE: this includes both active and non-active fault
        locations. Active fault locations are those of faults seen for the
        first time in this tie, while non-active are those of faults already
        seen in a previous tie.
        """
        if (collapse):
            return self._num_fault_groups
        else:
            return self._num_fault_locs

    def num_active_fault_locs(self, collapse=False) -> int:
        """
        Return only the number of active fault locations (either groups or
        elements) in this tie. An active fault location is a fault location
        that belongs to a fault that is seen for the first time in this tie.
        """
        return len(self._all_fault_locs)

    @deprecated(version='2.4.0', reason='This function has been renamed to '
                '`Tie.active_fault_locations`.')
    def fault_groups(self, collapse=False) -> Dict[Spectrum.Element, Set[Any]]:
        """
        Return all active fault locations (either groups or elements) in
        this tie, with the faults they contain. Active fault locations are
        those of faults seen for the first time in this tie.
        """
        return self.active_fault_locations(collapse=collapse)

    @versionadded(version='2.4.0', reason='Renamed from `Tie.fault_groups`.')
    def active_fault_locations(self, collapse=False) \
            -> Dict[Spectrum.Element, Set[Any]]:
        """
        Return all active fault locations (either groups or elements) in
        this tie, with the faults they contain. Active fault locations are
        those of faults seen for the first time in this tie.
        """
        if (collapse):
            return self._all_fault_groups
        else:
            return self._all_fault_locs

    def expected_value(self, q, weffort: bool, collapse=False) -> float:
        """
        Calculates the expected value of the qth fault in this tie. The
        expected value can either be in terms of wasted effort (not
        counting fault inspection) or actual position in the ranking.
        """
        fs = self.fault_groups(collapse)
        dups = sorted(chain(*fs.values()))
        m = self.len(collapse)
        l = self.num_fault_locs(collapse)
        l_a = self.num_active_fault_locs(collapse)
        nf = self.num_faults()
        if (weffort):
            expval = Fraction(m-l, l_a+1)
        else:
            expval = Fraction(m+1, l_a+1)
        # first fault or all 1 fault per loc
        if (q == 1 or (l_a == nf and all(len(ls) == 1 for ls in fs.values()))):
            return float(q*expval)
        elif (l_a > 10 or len(dups) == len(set(dups))):  # approx. solution or all single loc
            fpl = Fraction(nf, l_a)  # faults per (active) location
            for i in range(1, l_a+1):
                if (ceil(fpl*i) >= q):
                    if (fpl*i >= q):  # exactly contained in location
                        return float(expval*i)
                    else:  # contained between this location and the next
                        return float(expval*(i + (1 - ((fpl*i) % 1))))
            raise ValueError(f"Could not find fault {q}")
        else:
            dist = {}
            for rank in permutations(fs.keys()):
                seen = set()
                tot = 0
                for elem in rank:
                    tot += 1
                    seen.update(fs[elem])
                    if (len(seen) >= q):
                        break
                if (tot not in dist):
                    dist[tot] = 0
                dist[tot] += 1
            frac = Fraction(sum(v*c for v, c in dist.items()), factorial(l_a))
            return float(expval*frac)

    def _add_entity(self, entity: Spectrum.Entity,
                    faults: Dict[Any, Set[Spectrum.Element]],
                    seen_faults: Set[Any]):
        self._elems.update(entity)
        self._group_len += 1
        # Check if fault is in group
        faulty_group = False
        seen_fault_locs: Set[Spectrum.Element] = set()
        for (fault_num, fault_locs) in faults.items():
            unseen_fault = False
            for fault_loc in set(fault_locs).intersection(entity):
                # Check & update fault number
                if (fault_num not in seen_faults):
                    unseen_fault = True
                    self._num_faults += 1
                    seen_faults.add(fault_num)
                # Check & update faulty locs
                if (fault_loc not in seen_fault_locs):
                    seen_fault_locs.add(fault_loc)
                    self._num_fault_locs += 1
                    if (unseen_fault):
                        self._all_fault_locs.setdefault(fault_loc,
                                                        set()).add(fault_num)
                        self._fault_locs.setdefault(fault_num,
                                                    set()).add(fault_loc)
                # Check & update faulty group
                if (not faulty_group):
                    self._num_fault_groups += 1
                    faulty_group = True
                    if (unseen_fault):
                        self._all_fault_groups.setdefault(entity[0],
                                                          set()).add(fault_num)
                        self._fault_groups.setdefault(fault_num,
                                                      set()).add(entity[0])


class Ties:
    def __init__(self, rankings: Rankings):
        self.faults = rankings.faults()
        seen_faults: Set[Any] = set()
        seen_entities: Set[Spectrum.Entity] = set()
        self.ties: List[Tie] = []
        class RankingIter:
            def __init__(self, ranking: Ranking):
                self.r_iter = iter(ranking)
                self.cur: Optional[Rank] = next(self.r_iter, None)
            def is_active(self) -> bool:
                return self.cur is not None
            def consume(self) -> Rank:
                if (self.cur is None):
                    raise StopIteration("No more elements in ScoreIter")
                old_cur = self.cur
                self.cur = next(self.r_iter, None)
                return old_cur
            def cur_score(self):
                if (self.cur is not None):
                    return self.cur.score
                else:
                    raise StopIteration()
        r_iters = [RankingIter(r) for r in rankings]
        # Populate this Ties object
        def get_tie_entities(ri: RankingIter) -> Set[Spectrum.Entity]:
            entities: Set[Spectrum.Entity] = set()
            # sanity check
            if (not ri.is_active()):
                return entities
            # Get all UUTs with same score
            score = ri.cur_score()
            while (ri.is_active() and ri.cur_score() == score):
                r = ri.consume()
                entities.add(r.entity)
            return entities
        while (any(ri.is_active() for ri in r_iters)):
            all_entities: Set[Spectrum.Entity] = set()
            for ri in r_iters:
                if (ri.is_active()):
                    entities = get_tie_entities(ri)
                    all_entities.update(entities)
            tie = Tie()
            for entity in all_entities.difference(seen_entities):
                tie._add_entity(entity, self.faults, seen_faults)
            seen_entities.update(all_entities)
            self.ties.append(tie)
        # Add elements not seen to bottom of tie
        not_seen: Set[Spectrum.Element] = set(rankings.elements())
        for entity in seen_entities:
            not_seen.difference_update(entity)
        tie = Tie()
        for entity in not_seen:
            tie._add_entity(entity, self.faults, seen_faults)
        self.ties.append(tie)
        self._num_entities = len(seen_entities)
        self._num_elems = len(rankings.elements())

    def __iter__(self):
        return iter(self.ties)

    def size(self, collapse=False):
        if (collapse):
            return self._num_entities
        else:
            return self._num_elems



from __future__ import annotations
from typing import List, Dict, Any, Set, Sequence, Tuple, Callable, \
        Union, Iterable, Iterator, Optional, TYPE_CHECKING
from bitarray import bitarray
import numpy as np
from enum import Enum
from abc import ABC, abstractmethod
if TYPE_CHECKING:
    from flitsr.input import InputType

class Outcome(Enum):
    """ The outcome of a `Spectrum.Test` """
    PASSED = 0
    PASS = 0
    FAILED = 1
    FAIL = 1
    ERROR = 2


class Spectrum:
    """An implementation for a program spectrum."""
    class Test:
        """A test object holds information pertaining to a particular test."""

        __slots__ = ('name', 'index', 'outcome')

        def __init__(self, name: str, index: int, outcome: Outcome):
            self.name = name
            self.index = index
            self.outcome = outcome

        def __str__(self):
            return self.name

        def __repr__(self):
            return str(self)

        def __eq__(self, other):
            return self.name == other.name

        def __hash__(self):
            return hash(self.name)

    class Entity(ABC, Iterable):
        """A generic spectrum `Spectrum.Entity`. This is the base class for `Spectrum.Element`
        and `Spectrum.Group` in the spectrum.
        """
        __slots__ = ()

        @abstractmethod
        def isFaulty(self) -> bool:
            """Returns whether or not this `Spectrum.Entity` pertains to a fault"""
            pass

        @abstractmethod
        def index(self) -> int:
            """The index of this `Spectrum.Entity`"""
            pass

        @abstractmethod
        def __len__(self) -> int:
            pass

        @abstractmethod
        def __iter__(self) -> Iterator:
            pass

        @abstractmethod
        def __getitem__(self, index: int) -> Spectrum.Element:
            pass

    class Element(Entity):
        """
        An element object holds information pertaining to a single spectral
        element (line, method, class, etc...).
        """
        __slots__ = ('_index', '_group', 'path', 'method', 'line', 'faults',
                     'tup', 'hash')

        def __init__(self, details: List[str], index: int, faults: List[Any]):
            if (len(details) < 1):
                raise ValueError("Unnamed element: ", *details)
            self._index = index
            self.path = details[0]
            self.method = None
            self.line = None
            if (len(details) > 2):
                if (details[2].isdigit()):
                    self.line = int(details[2])
                    self.method = details[1]
                else:
                    self.method = details[1] + "|" + details[2]
            elif (len(details) > 1):
                if (not details[1].isdigit()):
                    self.method = details[1]
                else:
                    self.line = int(details[1])
            self.faults = faults
            self.tup = (self.path, self.method, self.line)
            self.hash = hash(self.tup)

        def isFaulty(self) -> bool:
            """Returns whether or not this `Spectrum.Element` pertains to a fault"""
            return len(self.faults) > 0

        def index(self) -> int:
            """Returns the index of this `Spectrum.Element`"""
            return self._index

        def __getitem__(self, index: int):
            if (index == 0):
                return self
            else:
                raise IndexError("list index out of range")

        def __iter__(self) -> Iterator:
            return iter((self,))

        def __len__(self) -> int:
            return 1

        def __str__(self):
            return "|".join(str(i) for i in self.tup if i) + \
                   (" (FAULT {})".format(",".join(str(x) for x in self.faults))
                    if self.faults else "")

        def output_str(self, type_: 'InputType', incl_faults=True) -> str:
            """
            Returns the string representation of this `Spectrum.Element` that can be
            used when writing out the spectrum in the given input type.

            Args:
              type_: 'InputType': The input type to render the output string for.
              incl_faults:  (Default value = True) Whether to include fault
                indicators in the output string

            Returns:
                The string representation of this `Spectrum.Element`

            """
            seps = type_.value.get_elem_separators()
            gstring = ''
            path_part = self.path.rpartition('.')
            if (path_part[0] != '' and path_part[2] != ''):
                gstring = path_part[0] + seps[0] + path_part[2]
            elif (path_part[0] != '' or path_part[2] != ''):
                gstring = path_part[0] + path_part[2]
            if (self.method):
                gstring += (seps[1] if (gstring != '') else '') + self.method
            if (self.line):
                gstring += (seps[2] if (gstring != '') else '') + str(self.line)
            if (incl_faults and self.isFaulty()):
                gstring += seps[3] + seps[3].join(str(x) for x in self.faults)
            return gstring

        def __repr__(self):
            return str(self)

        def __eq__(self, other):
            return (isinstance(other, Spectrum.Element) and
                    self.tup == other.tup)

        def __hash__(self):
            return self.hash

        def semantic_eq(self, other) -> bool:
            """
            Variant of the __eq__ method for an `Spectrum.Element` that checks whether
            two elements are semantically equivalent.

            Args:
              other: The element to compare to

            Returns:
              True if this `Spectrum.Element` is equal to `other`, False otherwise.

            """
            return self.tup == other.tup

    class Group(Entity):
        """
        A spectral group is a collection of spectral elements which have
        identical spectra.
        """
        __slots__ = ('_elems', '_is_faulty', '_index', '_elem_set', '_hash')

        def __init__(self, elems: Optional[List[Spectrum.Element]] = None,
                     index: Optional[int] = None):
            """
            Create a group object. If elems is given, the group is initialized
            with the elements within, otherwise an empty group is initialized.
            """
            self._elems = elems if elems is not None else []
            self._is_faulty = any(e.isFaulty() for e in self._elems)
            self._index = index
            self._elem_set = frozenset(self._elems)
            self._hash = hash(self._elem_set)

        def sort_elems(self, key: Callable):
            """
            Sort the elements in this group by the given `key`.
            Args:
              key: Callable: The key to sort by
            """
            self._elems.sort(key=key)

        def set_index(self, index: int):
            """
            Set the index of this group. Used for spectrum execution lookup.
            NOTE: the group index MUST be unique for each group in a given
            spectrum.
            Args:
              index: int: The integer
            """
            self._index = index

        def index(self) -> int:
            """
            Get the index of this Group if available, otherwise raise an
            ``UnboundLocalError``.
            """
            if (self._index is None):
                raise UnboundLocalError("Group index has not been set")
            return self._index

        def isFaulty(self) -> bool:
            """
            Returns whether or not an element in this `Spectrum.Group` pertains to a
            fault
            """
            return self._is_faulty

        def is_subgroup(self, group: Spectrum.Group) -> bool:
            """
            Return whether the given `group` is a subgroup of this `Spectrum.Group`.
            Args:
              group: Spectrum.Group: The group to check.

            Returns:
              True if `group` is a subgroup of this `Spectrum.Group`, False otherwise.
            """
            return group._elem_set < self._elem_set

        def __getitem__(self, index: int) -> Spectrum.Element:
            return self._elems[index]

        def __len__(self):
            return len(self._elems)

        def __iter__(self):
            return iter(self._elems)

        def __str__(self) -> str:
            return f"G{self._index} ({self._elems})"

        def __repr__(self) -> str:
            return str(self)

        def __eq__(self, other):
            return (isinstance(other, Spectrum.Group) and
                    self._elem_set == other._elem_set)

        def __hash__(self):
            return self._hash

    class Execution():
        """The Execution object holds all of the spectral information pertaining
        to the execution of a particular test.
        """
        __slots__ = ('_groups', '_spectrum', 'exec', 'test', '_iter')

        def __init__(self, test: Spectrum.Test,
                     spectrum: Spectrum):
            self._groups = spectrum.groups()
            self._spectrum = spectrum
            self.exec = bitarray(len(self._groups))
            self.test = test

        def update(self, group: Spectrum.Group, executed: bool):
            """Update the execution information of a particular group

            Args:
              group: Spectrum.Group: The group to update information for
              executed: bool: The updated execution information; whether this
                  group is executed or not.
            """
            self.exec[group.index()] = executed

        def __len__(self):
            return len(self._groups)

        def __iter__(self):
            self._iter = iter(self._groups)
            return self

        def __next__(self):
            return self.get(next(self._iter))

        def __getitem__(self, elem: Spectrum.Entity) -> bool:
            if (isinstance(elem, Spectrum.Group)):
                return self.get(elem, False)
            elif (isinstance(elem, Spectrum.Element)):
                return self.element_get(elem, False)
            else:
                raise TypeError(f"Illegal Entity type: {type(elem).__name__}")

        def __setitem__(self, group: Spectrum.Group, val: bool):
            try:
                self.exec[group.index()] = val
            except KeyError:
                pass

        def get(self, group: Spectrum.Group, default: bool = False) -> bool:
            """
            Return the execution for the given `group` if it is available, else
            default.

            Args:
              group: Spectrum.Group: The `Spectrum.Group` to lookup.
              default: bool:  (Default value = False) The default value if
                `group` has no execution information in this `Spectrum.Execution`.

            Returns:
              The execution information for `group` if `group` is found,
              `default` otherwise.
            """
            try:
                return bool(self.exec[group.index()])
            except (KeyError, IndexError):
                return default

        def element_get(self, elem: Spectrum.Element,
                        default: bool = False) -> bool:
            """
            Similar to the `get` method, but takes an `Spectrum.Element` instead of a
            `Spectrum.Group`.

            Args:
              elem: Spectrum.Element: The `Spectrum.Element` to lookup.
              default: bool:  (Default value = False) The default value if
              `elem` has no execution information in this `Spectrum.Execution`.

            Returns:
              The execution information for `elem` if the `Spectrum.Group` of `elem` is
              found, `default` otherwise.
            """
            # Get the element's group
            try:
                return bool(self.exec[self._spectrum.get_group(elem).index()])
            except (KeyError, IndexError):
                return default

    def __init__(self, elements: List[Spectrum.Element],
                 groups: List[Spectrum.Group], tests: List[Spectrum.Test],
                 executions: Dict[Test, Set[Spectrum.Element]]):
        self.spectrum: Dict[Spectrum.Test, Spectrum.Execution] = {}
        self.p: Dict[Spectrum.Group, int] = dict()
        self.f: Dict[Spectrum.Group, int] = dict()
        # Initialize element related properties
        edict = {e: i for i, e in enumerate(elements)}
        for group in groups:
            group.sort_elems(key=lambda x: edict[x])
        self._groups = sorted(groups, key=lambda g: edict[g[0]])
        self._group_map: Dict[Spectrum.Element, Spectrum.Group] = {}
        for i, group in enumerate(self._groups):
            group.set_index(i)
            self.p[group] = 0
            self.f[group] = 0
            for elem in group:
                self._group_map[elem] = group
        self._elements: List[Spectrum.Element] = elements
        # Initialize test related properties
        self._removed_tests: Dict[str, List[Spectrum.Test]] = {}
        self._removed_groups: Dict[str, List[Spectrum.Group]] = {}
        self._tests: List[Spectrum.Test] = tests
        self._failing: List[Spectrum.Test] = []
        self.tp: int = 0
        self.tf: int = 0
        for test in tests:
            # Increment total counts & add to failing set
            if (test.outcome is Outcome.PASSED):
                self.tp += 1
            else:
                self._failing.append(test)
                self.tf += 1
            # Add test execution
            self.spectrum[test] = self.Execution(test, self)
        # Initialize execution information
        for test in self._tests:
            test_exe = executions[test]
            seen: Set[Spectrum.Group] = set()
            for elem in test_exe:
                group = self._group_map[elem]
                if (group not in seen):
                    seen.add(group)
                    self.spectrum[test].update(group, True)
                    if (test.outcome is Outcome.PASSED):
                        self.p[group] += 1
                    else:
                        self.f[group] += 1

    def __getitem__(self, t: Test):
        return self.spectrum[t]

    def __iter__(self):
        self._iter = iter(self._tests)
        return self

    def __next__(self):
        return self.spectrum[next(self._iter)]

    def __len__(self):
        return len(self._tests)

    def elements(self) -> List[Spectrum.Element]:
        """Get a list of a all the elements in this spectrum"""
        return self._elements

    def groups(self) -> List[Spectrum.Group]:
        """Get a list of a all the groups in this spectrum"""
        return self._groups

    def tests(self) -> List[Spectrum.Test]:
        """Get a list of a all the tests (passing and failing) in this spectrum"""
        return self._tests

    def failing(self) -> List[Spectrum.Test]:
        """Get a list of a all the failing tests in this spectrum"""
        return self._failing

    def locs(self) -> int:
        """Get the number of groups in this spectrum"""
        return len(self._groups)

    def remove_test(self, test: Spectrum.Test, bucket='default'):
        """
        Remove the given test from this spectrum.
        All removed tests are stored so that they may be added back to the
        spectrum at a later point. By default, removed tests are stored in the
        'default' bucket, but a different bucket name may be optionally given.
        Setting the bucket to be None will completely remove the test from the
        spectrum (NOTE: this does not save much memory, but only prevents the
        test from being restored by subsequent calls to reset).

        Args:
          test: Spectrum.Test: The `Spectrum.Test` to remove.
          bucket:  (Default value = 'default') The bucket to store the removed
            tests in.
        """
        if (test not in self._tests):
            return
        self._tests.remove(test)
        if (test.outcome is Outcome.PASSED):
            self.tp -= 1
        else:
            self._failing.remove(test)
            self.tf -= 1
        for group in self.groups():
            self.remove_execution(test, group, hard=False)
        if (bucket is not None):
            self._removed_tests.setdefault(bucket, []).append(test)

    def remove_execution(self, test: Spectrum.Test, ent: Spectrum.Entity,
                         hard=True):
        """
        Remove the execution of the given `Spectrum.Entity` in `test`.

        Args:
          test: Spectrum.Test: The test to remove the execution of `ent` from.
          ent: Spectrum.Entity: The `Spectrum.Entity` whose execution to remove.
          hard:  (Default value = True) Whether to remove this execution fully,
            i.e. even from the underlying spectrum
        """
        if (self.spectrum[test][ent]):
            if (isinstance(ent, Spectrum.Element)):
                group = self._group_map[ent]
            elif (isinstance(ent, Spectrum.Group)):
                group = ent
            if (hard):
                self.spectrum[test][group] = False
            if (test.outcome is Outcome.PASSED):
                self.p[group] -= 1
            else:
                self.f[group] -= 1

    def remove_group(self, group: Spectrum.Group, bucket='default'):
        """
        Remove the given `group` from the spectrum, and store it in the
        given `bucket`. See `Spectrum.remove_test` for a description of buckets.

        Args:
          group: Spectrum.Group: The `Spectrum.Group` to remove.
          bucket:  (Default value = 'default') The bucket to store the group in
            for later retrieval.
        """
        try:
            self._groups.remove(group)  # remove if there
            self._removed_groups.setdefault(bucket, []).append(group)
        except ValueError:
            # Ignore if the element is not in the list
            pass
        self.p.pop(group, None)  # remove if there
        self.f.pop(group, None)  # remove if there

    def reset(self, bucket: Union[None, str, List[str]] = None):
        """
        Re-activates removed tests and recomputes counts.
        By default, or if bucket is None, all buckets are emptied and their
        removed tests reactivated. A bucket name (or Iterable of bucket names)
        may optionally be given to only reset the given name(s).

        Args:
          bucket: Union[None: str: List[str]]:  (Default value = None) The
              bucket to reset.
        """
        # first get the groups to add back
        groups_add_back = []
        group_keys_remove = []
        for key in self._removed_groups:
            if (bucket is None or (isinstance(bucket, str) and key == bucket)
               or (isinstance(bucket, List) and key in bucket)):
                groups_add_back.extend(self._removed_groups[key])
                group_keys_remove.append(key)
        for key in group_keys_remove:
            del self._removed_groups[key]
        # then add back the groups, re-computing counts
        for group in groups_add_back:
            self._groups.insert(group.index(), group)
            self.p[group] = len([t for t in self.tests() if t.outcome ==
                                Outcome.PASS and self[t][group]])
            self.f[group] = len([t for t in self.failing() if self[t][group]])
        # next get the tests to add back
        tests_add_back = []
        test_keys_remove = []
        for key in self._removed_tests:
            if (bucket is None or (isinstance(bucket, str) and key == bucket)
               or (isinstance(bucket, List) and key in bucket)):
                tests_add_back.extend(self._removed_tests[key])
                test_keys_remove.append(key)
        for key in test_keys_remove:
            del self._removed_tests[key]
        # then add back the tests, re-computing counts
        for test in tests_add_back:
            self._add_back_removed_test(test)

    def _add_back_removed_test(self, test: Spectrum.Test):
        """
        Helper method to add back tests that have been removed from the
        spectrum.
        """
        self._tests.append(test)
        if (test.outcome is Outcome.PASSED):
            self.tp += 1
        else:
            self._failing.append(test)
            self.tf += 1
        for group in self.groups():
            if (self.spectrum[test][group]):
                if (test.outcome is Outcome.PASSED):
                    self.p[group] += 1
                else:
                    self.f[group] += 1

    def reset_single_test(self, test: Spectrum.Test,
                          bucket: Optional[str] = None):
        """
        Reset a single test that was stored in the given bucket. See the
        `Spectrum.reset` method, which is the iterative form of this one.

        Args:
          test: Spectrum.Test: The `Spectrum.Test` to reset.
          bucket: Optional[str]:  (Default value = None) The bucket that `test`
            is in, or ``None`` to check all buckets.

        Raises:
          KeyError: If the test could not be found in the given bucket.
        """
        if (bucket is None):
            for key in self._removed_tests:
                if (test in self._removed_tests[key]):
                    bucket = key
                    break
        if (bucket is None or test not in self._removed_tests[bucket]):
            raise KeyError(f"Could not find removed test {test} in "
                           f"{bucket or 'any'} bucket")
        self._removed_tests[bucket].remove(test)
        self._add_back_removed_test(test)

    def get_group(self, element: Spectrum.Element) -> Spectrum.Group:
        """
        Given an element, return the `Spectrum.Group` with identical coverage
        from this spectrum that this element is apart of.

        Args:
          element: Spectrum.Element: The `Spectrum.Element` to find the
            `Spectrum.Group` of.

        Returns:
          The group that the `element` belongs to.

        Raises:
          KeyError: If the element does not belong to any group in this
            spectrum.
        """
        return self._group_map[element]

    def get_faults(self) -> Dict[Any, Set[Spectrum.Element]]:
        """
        Returns a dictionary of all the faults in this spectrum, with the
        values being all the locations of each fault.
        """
        actual_faults: Dict[Any, Set[Spectrum.Element]] = dict()
        for group in self.groups():
            for elem in group:
                if (elem.faults):
                    for fault in elem.faults:
                        actual_faults.setdefault(fault, set()).add(elem)
        return actual_faults

    def get_tests(self, entity: Spectrum.Entity, only_failing=False,
                  remove=False, bucket='default') -> Set[Spectrum.Test]:
        """
        Finds all the test cases executing the given element, and (optionally)
        removes them from the spectrum. If removing tests from the spectrum, an
        optional bucket name may also be given.

        Args:
          entity: Spectrum.Entity: The entity to find tests for.
          only_failing:  (Default value = False) Whether to return only tests
              that are failing.
          remove:  (Default value = False) Whether to remove the found tests
              from the spectrum before returning them.
          bucket:  (Default value = 'default') The bucket to add the tests to
              if they are removed from the spectrum.

        Returns:
            The set of `Spectrum.Test` that `entity` is executed in.

        """
        executing = set()
        tests = self.failing() if only_failing else self.tests()
        for test in tests:
            if (self[test][entity]):
                executing.add(test)
        if (remove):
            for test in executing:
                self.remove_test(test, bucket=bucket)
        return executing

    def get_executed_groups(self, test: Spectrum.Test) -> Set[Spectrum.Group]:
        """
        Return the set of groups that are executed in the given `test`.

        Args:
          test: Spectrum.Test: The test whose executed groups to return.

        Returns:
          The set of `Spectrum.Group` that are executed in the given `test`.

        """
        return self._get_executed_entities(test, groups=True)

    def get_executed_elements(self, test: Spectrum.Test) -> Set[Spectrum.Element]:
        """
        Return the set of elements that are executed in the given `test`.

        Args:
          test: Spectrum.Test: The test whose executed elements to return.

        Returns:
          The set of `Spectrum.Element` that are executed in the given `test`.

        """
        return self._get_executed_entities(test, groups=False)

    def get_removed_tests(self, bucket='default') -> List[Spectrum.Test]:
        """
        Return a copy of all the removed tests in a particular bucket. If
        bucket is None, all removed tests from all buckets are returned.

        Args:
          bucket:  (Default value = 'default') The bucket to look for removed
            tests to return.

        Returns:
          A `List` of all the `Spectrum.Test` that were in the given bucket, or
          in all buckets if `bucket` is None.
        """
        if (bucket is None):
            return self._all_removed_tests()
        else:
            return self._removed_tests[bucket][:]

    def _get_executed_entities(self, test: Spectrum.Test, groups=True):
        """
        Finds either the groups (default) or the elements (groups = False)
        executed in the given test.
        """
        executed = set()
        entities: Sequence[Spectrum.Entity]
        if (groups):
            entities = self._groups
        else:
            entities = self._elements
        for ent in entities:
            if (self[test][ent]):
                executed.add(ent)
        return executed

    def _all_removed_tests(self) -> List[Spectrum.Test]:
        """ Get a list of the removed tests in all buckets """
        all_removed = []
        for key in self._removed_tests:
            all_removed.extend(self._removed_tests[key])
        return all_removed

    def to_matrix(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Converts the current spectrum into a numpy matrix and error vector.

        Returns:
          A numpy matrix representing the spectrum, and numpy array
          representing the error vector.
        """
        # If no matrix already, create one
        if (not hasattr(self, '_matrix')):
            # Use all test cases and most elements
            self._matrix_tests = list(self.spectrum.keys())
            self._matrix_elems = list(self._groups)
            self._matrix = np.zeros((len(self._matrix_tests),
                                     len(self._matrix_elems)),
                                    dtype=bool)
            self._errVector = np.zeros(len(self._matrix_tests),
                                       dtype=bool)
            for (i, test) in enumerate(self._matrix_tests):
                for (j, elem) in enumerate(self._matrix_elems):
                    self._matrix[i][j] = self.spectrum[test][elem]
                self._errVector[i] = 0 if (test.outcome is
                                           Outcome.PASSED) else 1
        # Extract submatrix
        test_set = set(self._tests)
        tmask = np.array([item in test_set for item in self._matrix_tests])
        # tmask = np.isin(self._matrix_tests, self._tests)
        group_set = set(self._groups)
        emask = np.array([item in group_set for item in self._matrix_elems])
        # emask = np.isin(self._matrix_elems, self._groups)
        matrix = self._matrix[np.ix_(tmask, emask)]
        errVector = self._errVector[tmask]
        return matrix, errVector

    def search_tests(self, name_part,
                     incl_removed=False) -> List[Spectrum.Test]:
        """
        Searches for any test that matches the partial test name `name_part`.
        Returns a list of all the matches.

        Args:
          name_part: The partial name of a test to search for.
          incl_removed:  (Default value = False) Whether to include tests that
            have been removed.

        Returns:
          A list of all the tests that match `name_part`.
        """
        results = [t for t in self.tests() if t.name.find(name_part) != -1]
        if (incl_removed):
            results.extend([t for t in self._all_removed_tests()
                            if t.name.find(name_part) != -1])
        return results

    def search_elements(self, name_part, groups=False) \
            -> List[Union[Spectrum.Element, Spectrum.Group]]:
        """
        Searches for any element that matches the partial element name
        `name_part`. Returns a list of all the matches.

        Args:
          name_part: The partial name of an entity to search for.
          groups:  (Default value = False) Whether to search for
            `Spectrum.Group` or `Spectrum.Element`.

        Returns:
          A list of all the entitied that match `name_part`.
        """
        elems = self._groups if groups else self._elements
        results = [e for e in elems if str(e).find(name_part) != -1]
        return results



from __future__ import annotations
import argparse
import argcomplete
import inspect
import sys
import re
from warnings import warn
from pathlib import Path
from os import path as osp
from typing import List, Dict, Any, Optional, IO, BinaryIO, Tuple, Set, \
        Union, Generic
from flitsr.suspicious import Suspicious
from flitsr import cutoff_points
from flitsr.singleton import SingletonMeta
from flitsr.ranking import Tiebrk
from flitsr import advanced
from flitsr.advanced import Config


class Args(argparse.Namespace, metaclass=SingletonMeta):
    """
    The global Singleton holding the specified/default arguments used across
    the flitsr framework.
    """
    def __init__(self, argv: Optional[List[str]] = None, cmd_line=False,
                 noparse=False):
        """
        Initialize the Args Singleton object. Note that this constructor is
        only ever called once for each program run. Other attempts to construct
        this object will return the pre-existing ``Args`` Singleton object.

        Args:
          argv: Optional[List[str]]: The optional list of command line
              arguments to parse. If not given and the Args object is set to
              read from the command line (see ``cmd_line``), sys.argv is used
              instead.
          cmd_line: bool: Whether to parse arguments given in command-line
              style. When True, Args will either parse the arguments given in
              ``argv``, or ``sys.argv`` if ``argv`` is ``None``. If False,
              ``Args`` will just load the default values for each command line
              option, and ignore required arguments. (Default value = False)
          noparse: bool: When True, this ``Args`` object will not attempt to
              parse any arguments. This is primarily used for generating
              documentation for flitsr's command line. (Default value = False)
        """
        self._deprecations: Dict[str, str] = {}
        self._default_metric = 'ochiai'
        self._adv_required_args: Dict[str, Tuple[str, Set[str]]] = dict()
        self._advanced_params: Dict[str, Dict[str, Any]] = {}
        self._parser = self._gen_parser(cmd_line)
        if (not noparse):
            self._parse_args(argv)

    def _add_params(self, params: argparse.Namespace):
        dict_ = vars(params)
        for key in dict_:
            setattr(self, key, dict_[key])

    @staticmethod
    def _check_file(filename):
        """ Check file exists function """
        if (osp.exists(filename)):
            return filename
        else:
            raise argparse.ArgumentTypeError('Could not find input file:'
                                             f' \"{filename}\"')

    def _check_type(self, type_comb: str):
        """ check FLITSR type combinations function """
        type_sep = type_comb.split('+')
        cluster = None
        ranker = None
        refiner = None
        args: Dict[str, Dict[str, Any]] = {}
        for t in type_sep:
            m = re.fullmatch("([^(]+)(?:\\(([^)]+)\\))?", t)
            if (m is None):
                raise argparse.ArgumentTypeError('Invalid type for '
                                                 f'--all-types: \"{t}\"')
            t = m.group(1).upper()
            if (hasattr(advanced.ClusterType, t)):
                if (cluster is not None):
                    raise argparse.ArgumentTypeError('Cannot have two '
                        f'cluster types: {cluster.name} and {t}')
                cluster = advanced.ClusterType[t]
            elif (hasattr(advanced.RefinerType, t)):
                if (refiner is not None):
                    raise argparse.ArgumentTypeError('Cannot have two '
                        f'refiner types: {refiner.name} and {t}')
                refiner = advanced.RefinerType[t]
            elif (hasattr(advanced.RankerType, t)):
                if (ranker is not None):
                    raise argparse.ArgumentTypeError('Cannot have two '
                        f'ranker types: {ranker.name} and {t}')
                ranker = advanced.RankerType[t]
            else:
                raise argparse.ArgumentTypeError('Invalid type for '
                                                 f'--all-types: \"{t}\"')
            if (m.group(2) is None):
                params = {}
            else:
                params = dict(re.findall("([^=,]+)=['\"]?([^,'\"]+)['\"]?",
                                         m.group(2)))
                # check params are valid and convert types
                for k, v in params.items():
                    # check valid param
                    if (k not in self._advanced_params[t]):
                        raise argparse.ArgumentTypeError('Invalid parameter '
                                                         f'\"{k}\" for type '
                                                         f'{t}')
                    else:
                        # convert type
                        try:
                            converted = self._advanced_params[t][k](v)
                            params[k] = converted
                        except Exception:
                            raise argparse.ArgumentTypeError('Invalid value '
                                f'\"{v}\" for parameter {k} of type {t}')
            args[t] = params
        adv_type = Config(ranker, cluster, refiner, args)
        return adv_type

    @staticmethod
    def _get_base_type(typ) -> List[type]:
        """ function to extract the base type from a typing.Union, etc."""
        if (hasattr(typ, '__origin__')):
            if (typ.__origin__ is Union):
                ret = list(typ.__args__)
                # remove NoneType from Optionals
                if (type(None) in ret):
                    ret.remove(type(None))
                return ret
            elif (typ.__origin__ is Generic):
                return [typ.__args__[0]]
        return [typ]

    def _gen_parser(self, cmd_line: bool = True) -> argparse.ArgumentParser:
        """
        Generate the ArgumentParser that will be used to parse arguments.
        This function is used by the ``parse_args`` function, as well as to
        initialize the Args object and generate documentation. Setting
        ``cmd_line`` to False will disable required arguments, allowing the
        default values to be retrived.

        Args:
          cmd_line: bool: Whether to generate the parser for the command line,
              or just for the default arguments. (Default value = True)

        Returns:
            The argparse ArgumentParser that was generated.
        """
        # General options
        parser = argparse.ArgumentParser(prog='flitsr', description='An automatic '
                'fault finding/localization tool for multiple faults.')
        if (cmd_line):
            parser.add_argument('input', type=Args._check_file,
                                help='The coverage file (TCM) or directory '
                                '(GZoltar) containing the coverage collected '
                                'for the system over the test suite')
        parser.add_argument('-o', '--output', action='store', default=sys.stdout,
                type=argparse.FileType('w'), help='Specify the output file to use '
                'for all output (default: STDOUT).')
        parser.add_argument('--csv', action='store_true',
                help='By default FLITSR will output the ranking in it\'s own '
                'FLITSR ranking format. Enabling this option will allow FLITSR to '
                'output the ranking in CSV format compatible with GZoltar\'s CSV '
                'ranking format instead.')
        parser.add_argument('--spectrum-csv', action='store_true',
                help='Enabling this option will cause FLITSR to '
                'output the spectrum in CSV format.')
        metric_choices = (Suspicious.getNames(True) +
                          [m.lower() for m in advanced.all_types.keys()])
        parser.add_argument('-m', '--metric', dest='metrics', action='append',
                choices=metric_choices, metavar='METRIC',
                help='The underlying (SBFL) metric(s) to use when either ranking '
                '(if the sbfl option is given), or running the FLITSR algorithm. '
                'Option may be supplied multiple times to run multiple metrics. '
                'Specifying multiple metrics will output the results of each '
                'metric to a seperate file using the metric\'s name instead of '
                f'stdout. Allowed values are: [{", ".join(Suspicious.getNames(True))}] '
                f'(default: {self._default_metric}). When using the FLITSR '
                'advanced technique (or similar), you may also specify '
                'an advanced technique here to use as base metric, in which '
                'case the --flitsr-default_metric option may be used to '
                'likewise specify the base metric for that advanced '
                'technique. Note that it is your job to determine whether the '
                'given advanced technique is compatible with the FLITSR (or '
                'similar) algorithm, as this is not ensured for all advanced '
                'types.')
        parser.add_argument('-r', '--ranking', action='store_true',
                help='Changes FLITSR\'s expected input to be an SBFL ranking in '
                'Gzoltar or FLITSR format (determined automatically), instead of '
                'the usual coverage, and produces the specified calculations '
                '(or just the ranking if no calculations are given). NOTE: any '
                'non-output based options will be ignored with this option')
        parser.add_argument('-c', '--method', action='store_true',
                help='The default for FLITSR is to use the collected coverage '
                'as-is and merely produce the ranking in terms of the names/labels'
                ' given to the elements. Alternatively, using this option, FLITSR '
                'can assume the coverage given is a statement level coverage, and '
                'will attempt to collapse this coverage to produce a method level '
                'coverage result. This collapse is done by constructing a '
                'coverage matrix with only the method names, where the execution '
                'of a method is determined by the union of the executions of its '
                'statements. Bugs added to the coverage are handled in a similar '
                'fashion.')
        parser.add_argument('--split', help='When given, this option causes faults'
                ' that are a combination of two or more sub-faults in mutually'
                ' exclusive parts of the system to be split into separate'
                ' identified faults. As a by-product this also drops faults that'
                ' are not exposed by failing tests', action='store_true')
        parser.add_argument('--collapse', action='store_true',
                help='Collapse dynamic basic-block groups into singular elements '
                'for the ranking an calculations')
        parser.add_argument('-a', '--all', action='store_true',
                help='Used in the evaluation of FLITSR against other techniques. '
                'Runs all metrics given in suspicious.py and both FLITSR and '
                'FLITSR* extensions over each metric. Also enables all of the '
                'above evaluation calculations. Prints the results out to files '
                'named [<flitsr method>_]<metric>.run for each FLITSR method '
                'and metric')
        adv_types = list(advanced.all_types.keys())
        parser.add_argument('-t', '--types', action='append', type=self._check_type,
                             help='Specify the advanced type combination to use '
                             'when running FLITSR. Note that this argument '
                             'overrides any of the individual advanced type '
                             'arguments given to FLITSR (such as --multi, --sblf, etc.).'
                             'This argument may be specified multiple times to add '
                             'multiple type combinations to run. The format for '
                             'this argument is: "--types <type>[+<type>...]", '
                             'where each <type> is a (case-insensitive) FLITSR '
                             f'advanced type. Allowed types are: {adv_types}')
        parser.add_argument('-p', '--print-params', nargs='?', const="",
                            default=None, type=str, metavar='FORMAT_STR',
                            help='When producing multiple output files, '
                            'specifies that the parameters for advanced types,'
                            'given by the --types option, should be included '
                            'in the names of these output files. Without an '
                            'argument, this option produces file names with '
                            'advanced type parameters of the form "{k}-{v}" '
                            'where {k} is the parameter name, and {v} is the '
                            'value. Other formats can be constructed by '
                            'instead providing an argument using the same '
                            '"{k}", "{v}" syntax.')
        parser.add_argument('--no-override', action='store_true',
                            help='By default FLITSR will override the output '
                            'file(s) if they already exist, printing a warning '
                            'message. This option instead allows FLITSR to leave '
                            'output files that already exist, skipping that '
                            'output and continuing with the rest of the outputs '
                            '(if any)')
        parser.add_argument('-d', '--decimals', action='store', type=int,
                default=2, help='Sets the precision (number of decimal points) for '
                'the output of all of the calculations (default: %(default)s)')

        # Advanced types options
        primitives = (bool, str, int, float, Path)

        advanced_groups: List[Tuple[str, Any, Any, Any]] = []

        refiner_enum = advanced.RefinerType
        if (len(refiner_enum) != 0):
            refiners = parser.add_argument_group('Spectrum refiner techniques',
                'One of the following spectrum refining techniques may be '
                'specified, along with it\'s options')
            refiner_mu = refiners.add_mutually_exclusive_group()
            advanced_groups.append(('refiner', refiner_enum, refiners,
                                    refiner_mu))

        cluster_enum = advanced.ClusterType
        if (len(cluster_enum) != 0):
            clusters = parser.add_argument_group('Clustering techniques',
                'One of the following clustering techniques may be specified, '
                'along with it\'s options')
            cluster_mu = clusters.add_mutually_exclusive_group()
            advanced_groups.append(('cluster', cluster_enum, clusters,
                                    cluster_mu))

        ranker_enum = advanced.RankerType
        if (len(ranker_enum) != 0):
            rankers = parser.add_argument_group('Ranking techniques',
                'One of the following advanced ranking techniques may be '
                'specified, along with it\'s options (default: flitsr)')
            ranker_mu = rankers.add_mutually_exclusive_group()
            advanced_groups.append(('ranker', ranker_enum, rankers, ranker_mu))

        for adv_name, adv_enum, group, mu in advanced_groups:
            for type_ in list(adv_enum):
                name, class_ = type_.name, type_.value
                init = class_.__init__
                disp_name = name.lower()
                help_ = ''  # '(default: %(default)s)'
                # add docstring
                if (class_.__doc__ is not None and class_.__doc__ != ''):
                    help_ = (re.sub("\\s+", " ", class_.__doc__.strip()) +
                             ' ' + help_)
                # add cmd line argument for this advanced type
                mu.add_argument('--'+disp_name, dest=adv_name,
                                action='store_const', const=adv_enum[name],
                                help=help_)
                argspec = inspect.getfullargspec(init)
                def_diff = (len(argspec.args)-1) - (0 if
                                                    argspec.defaults is None
                                                    else len(argspec.defaults))
                # add cmd-line arguments for each parameter of this adv type
                self._advanced_params[name] = dict()
                for p_index, param in enumerate(argspec.args[1:]):
                    self._advanced_params[name][param] = None
                    # skip adding this parameter if it is marked as existing
                    if (hasattr(init, '__existing__') and
                        param in init.__existing__):
                        continue
                    paramName = '--'+disp_name+'-'+param.replace('_', '-')
                    parser_args: Dict[str, Any] = {}
                    # get the parameter type(s) (if any)
                    paramTypes = (self._get_base_type(argspec.annotations[param])
                                  if param in argspec.annotations else None)
                    # add default arguments if param has
                    if (argspec.defaults is not None and p_index >= def_diff):
                        parser_args['default'] = argspec.defaults[p_index-def_diff]
                        parser_args['help'] = '(default: %(default)s)'
                    # else add param as conditionally "required" if not bool
                    elif (paramTypes is None or paramTypes != [bool]):
                        if (name not in self._adv_required_args):
                            self._adv_required_args[name] = (adv_name, set())
                        self._adv_required_args[name][1].add(param)
                        parser_args['help'] = '(required)'
                    # add choices
                    if (hasattr(init, '__choices__') and
                        param in init.__choices__):
                        parser_args['choices'] = init.__choices__[param]
                    # add types
                    if (paramTypes is not None):
                        # function to deal with non-primitives
                        def non_primitive(name, param):
                            if (not hasattr(class_, f'_{param}')):
                                err = ('Could not find type conversion '
                                       f'for {param} in {name}')
                                raise NameError(err)
                            return getattr(class_, f'_{param}')

                        # only add automatic type conversion for single types
                        if (len(paramTypes) == 1):
                            paramType = paramTypes[0]
                            # specially handle bools
                            if (paramType is bool):
                                if ('default' in parser_args):
                                    # check if default is True
                                    if (parser_args['default']):
                                        parser_args['dest'] = disp_name+'_'+param
                                        paramName = ('--' + disp_name + '-no-' +
                                                     param.replace('_', '-'))
                                        parser_args['action'] = 'store_false'
                                    else:
                                        parser_args['action'] = 'store_true'
                                else:
                                    parser_args['action'] = 'store_true'
                            # specially handle file IO types
                            elif (inspect.isclass(paramType) and
                                  issubclass(paramType, IO)):
                                if (issubclass(paramType, BinaryIO)):
                                    parser_args['type'] = argparse.FileType('rb')
                                else:
                                    parser_args['type'] = \
                                      argparse.FileType('r', encoding='UTF-8')
                            else:
                                if (paramType not in primitives):
                                    paramType = non_primitive(name, param)
                                parser_args['type'] = paramType
                                if ('choices' not in parser_args):
                                    parser_args['metavar'] = param
                        # deal with multiple parameter types
                        else:
                            parser_args['type'] = non_primitive(name, param)
                            if ('choices' not in parser_args):
                                parser_args['metavar'] = param
                        # store parser type converter
                        if ('type' in parser_args):
                            self._advanced_params[name][param] = parser_args['type']
                        else:  # must be bool type
                            def boolConv(x: str) -> bool:
                                if (x not in ['True', 'False']):
                                    raise ValueError('invalid literal for '
                                                     f'bool: \'{x}\'')
                                return x == 'True'
                            self._advanced_params[name][param] = boolConv

                    # finalize option
                    group.add_argument(paramName, **parser_args)

        # parser.add_argument('--multi', action='store_true',
        #         help='Runs the FLITSR* (i.e. multi-round) algorithm')
        # parser.add_argument('-i', '--internal-ranking', action='store',
        #         choices=['flitsr', 'reverse', 'original', 'auto', 'conf'], default='auto',
        #         help='Specify the order in which the elements of each FLITSR basis '
        #         'are ranked. "flitsr" uses the order that FLITSR returns the basis '
        #         'in (i.e. from FLITSRs lowest to highest recursion depth), which '
        #         'aligns with FLITSRs confidence for each element being a fault. '
        #         '"reverse" uses the reverse of "flitsr" (i.e. the order in which '
        #         'FLITSR identifies the elements) which gives elements that use a '
        #         'larger part of the original test suite first. "original" returns '
        #         'the elements based on their original positions in the ranking '
        #         'produced by the base SBFL metric used by FLITSR.')

        # Tie breaking options
        tie_grp = parser.add_argument_group('Tie breaking strategy',
                                            'Specifies the tie breaking '
                                            'strategy to use for FLITSR and '
                                            'the localization')
        tie_grp.add_argument('--tiebrk', action='store_const',
                             default=Tiebrk.ORIG, const=Tiebrk.EXEC,
                             help='Breaks ties using only execution counts')
        tie_grp.add_argument('--rndm', action='store_const', const=Tiebrk.RNDM,
                             dest='tiebrk', help='Breaks ties by a randomly '
                             'ordering')
        tie_grp.add_argument('--otie', action='store_const', const=Tiebrk.ORIG,
                             dest='tiebrk', help='Breaks ties by using the '
                             'original base metric ranking (in the case of '
                             'FLITSR) and by execution counts otherwise')

        # Calculation options
        calc_grp = parser.add_argument_group('Calculations',
                'The following arguments replace the default ranking output of '
                'FLITSR with evaluation calculations. Multiple of the following '
                'arguments can be given in the same call')

        # check type of general weffort
        def check_fault_type(nth):
            if (int(nth) > 0):
                return int(nth)
            else:
                raise argparse.ArgumentTypeError(f'Invalid fault number {nth}')

        # Wasted effort calcuation options
        calc_grp.add_argument('--first', dest='weff', action='append_const',
                default=[], const='first',
                help='Display the wasted effort to the first fault')
        calc_grp.add_argument('--avg', '--average', dest='weff',
                action='append_const', const='avg',
                help='Display the wasted effort to the average fault')
        calc_grp.add_argument('--med', '--median', dest='weff',
                action='append_const', const='med',
                help='Display the wasted effort to the median fault')
        calc_grp.add_argument('--last', dest='weff', action='append_const',
                const='last', help='Display the wasted effort to the last fault')
        calc_grp.add_argument('--weffort', dest='weff', action='append', metavar='N',
                type=check_fault_type, help='Display the wasted effort to the Nth fault')

        # TOP1 calculation options
        def all_top(value):
            return ('all', pr(value))
        def one_top(value):
            return ('one', pr(value))
        def perc_top(value):
            return ('perc', pr(value))

        calc_grp.add_argument('--one-top', dest='top1', metavar='x',
                action='append', type=one_top, default=[],
                help='Display the expected value of finding at least one '
                'fault in the top `x` elements (elements with the '
                'highest suspiciousness).')
        calc_grp.add_argument('--all-top', dest='top1', metavar='x',
                action='append', type=all_top,
                help='Display the expected value of the total number of '
                'faults found in the top `x` elements (elements with the '
                'highest suspiciousness)')
        calc_grp.add_argument('--perc-top', dest='top1', metavar='x',
                action='append', type=perc_top,
                help='Display the expected value of the percentage of faults '
                'found in the top `x` elements (elements with the highest '
                'suspiciousness)')
        self._deprecations['--one-top1'] = ('`--one-top1` is deprecated and '
                                            'will be removed in an upcomming '
                                            'release. Use `--one-top` instead')
        calc_grp.add_argument('--one-top1', dest='top1',
                action='append_const', const=('one', 1),
                help='(Deprecated: Use `--one-top` instead) Display the '
                'expected value of finding at least one fault in the first '
                'element in the ranking.')
        self._deprecations['--all-top1'] = ('`--all-top1` is deprecated and '
                                            'will be removed in an upcomming '
                                            'release. Use `--all-top` instead')
        calc_grp.add_argument('--all-top1', dest='top1',
                action='append_const', const=('all', 1),
                help='(Deprecated: Use `--all-top` instead) Display the '
                'expected value of the total number of faults found in the '
                'first element in the ranking')
        self._deprecations['--perc-top1'] = ('`--perc-top1` is deprecated and '
                                            'will be removed in an upcomming '
                                            'release. Use `--perc-top` instead')
        calc_grp.add_argument('--perc-top1', dest='top1',
                action='append_const', const=('perc', 1),
                help='(Deprecated: Use `--perc-top` instead) Display the '
                'percentage of faults found in the first element in the '
                'ranking')

        # Percent at n calculation options
        calc_grp.add_argument('--perc@n', '--percent-at-n', dest='perc_at_n',
                action='append_const', const='perc', default=[],
                help='Produces the percentage-at-N values (i.e. the percentage of '
                'faults found at N%% of code inspected). The output of the perc@n '
                'calculation is a list of ranks of all found faults, preceded by '
                'the number of elements in the system, which can be used to '
                'generate percentage-at-N/recall graphs')
        calc_grp.add_argument('--auc', '--area-under-curve', dest='perc_at_n',
                action='append_const', const='auc',
                help='Dislpays the area under the curve produced by the '
                'percentage-at-N calculation')
        calc_grp.add_argument('--pauc', '--percent-area-under-curve',
                dest='perc_at_n', action='append_const', const='pauc',
                help='Dislpays the area under the curve '
                'produced by the percentage-at-N calculation as a percentage of '
                'the maximum possible value (i.e. closest to perfect recall)')
        calc_grp.add_argument('--lauc', '--log-area-under-curve',
                dest='perc_at_n', action='append_const', const='lauc',
                help='Dislpays the area under the curve '
                'produced by the percentage-at-N calculation as a logarithmic '
                'percentage of the maximum possible value (i.e. closest to perfect '
                'recall). The logarithmic effect causes lower ranks to have a '
                'greater effect on the value, which corresponds to the lower ranks '
                'being more useful to the developer')

        # Precision recall type functions
        def pr(value):
            if (value == "b" or value == "f"):
                return value
            elif (value.isdigit()):
                return int(value)
            else:
                raise ValueError(str(value)+" is not a valid precision/recall "
                                 "string value")
        def precision(value):
            return ('p', pr(value))
        def recall(value):
            return ('r', pr(value))

        # Precision recall calculation options
        calc_grp.add_argument('--precision-at', dest='prec_rec', metavar='x',
                action='append', default=[], type=precision,
                help='Displays precision values at a given rank `x`. Precision '
                'is the amount of faults f found within the cut-off point `x`, '
                'out of the number of elements seen (i.e. f/x). Can be specified '
                'multiple times')
        calc_grp.add_argument('--recall-at', dest='prec_rec', metavar='x',
                action='append', default=[], type=recall,
                help='Displays recall values at a given rank `x`. Recall '
                'is the amount of faults f found within the cut-off point `x`, '
                'out of the total number of faults n (i.e. f/n). Can be specified '
                'multiple times')

        # Fault output options
        calc_grp.add_argument('--fault-num', dest='faults', action='append_const',
                              const='num', default=[],
                              help='Display the number of faults in the program')
        calc_grp.add_argument('--fault-ids', dest='faults', action='append_const',
                              const='ids',
                              help='Display the IDs of the faults in the program')
        calc_grp.add_argument('--fault-elems', dest='faults',
                              action='append_const', const='elems',
                              help='Display the elements that are faulty in the program')
        calc_grp.add_argument('--fault-all', dest='faults', action='append_const',
                              const='all',
                              help='Display all info of the faults in the program')

        # parser.add_argument('-p', '--parallel', action='store',
        #          choices=parallel_opts, metavar='ALGORITHM',
        #          help='Run one of the parallel debugging algorithms on the spectrum '
        #          'to produce multiple spectrums, and process all other options on '
        #          'each spectrum. Allowed values are: ['+', '.join(parallel_opts)+']')

        # parser.add_argument('--artemis', action='store_true',
        #                     help='Run the ARTEMIS technique on the spectrum to '
        #                     'produce the ranked lists. This option may be '
        #                     'combined with FLITSR and parallel to produce a '
        #                     'hybrid technique.')

        # TODO: Add OBA and worst args if necessary
        cut_eval_opts=['worst', 'best', 'resolve']
        parser.add_argument('--cutoff-eval', action='store', metavar='MODE',
                choices=cut_eval_opts, help='Specifies the performance mode to use '
                'when using a multi-fault fixing cut-off strategy to produce '
                'rankings. Allowed values are: ['+', '.join(cut_eval_opts)+']'+
                ' (default %(default)s)', default='worst')
        cutoff_opts = cutoff_points.getNames()
        parser.add_argument('--cutoff-strategy', action='store', metavar='STRATEGY',
                help='Cuts off the ranking using the given '
                'strategy\'s cut-off point. This affects both the rank output '
                'method and any calculations. Allowed values are: ['+
                ', '.join(cutoff_opts)+'] (default %(default)s). '
                'For basis, an optional value n may be given (e.g. basis=n) '
                'that determines the number of bases included before the cutoff')

        argcomplete.autocomplete(parser)
        return parser

    def _parse_args(self, argv: Optional[List[str]] = None) -> Args:
        """
        Parse the arguments defined by args for the flitsr program with
        python's argparse. The result is an argparse Namespace object which
        includes all of the arguments parsed (or default values).

        Args:
          argv: Optional[List[str]]: (Default value = None) The list of
              arguments to parse, usual taken from the command line arguments
              given

        Returns:
          The constructed Args object
        """

        args = self._parser.parse_args(argv)

        # check for deprecations
        for dep, message in self._deprecations.items():
            if ((argv is not None and dep in argv) or
                    (argv is None and dep in sys.argv)):
                # stacklevel=10 to remove all context
                warn(message, category=FutureWarning, stacklevel=10)

        # manually set flitsr as the default
        default_ranker_used = False
        if (not hasattr(args, 'ranker') or args.ranker is None):
            args.ranker = advanced.RankerType['FLITSR']
            default_ranker_used = True

        # check "required" advanced type arguments
        for adv_name, (adv_type, adv_args) in self._adv_required_args.items():
            if (getattr(args, adv_type) is not None and
                getattr(args, adv_type).name == adv_name):
                dname = adv_name.lower()
                for adv_arg in adv_args:
                    if (getattr(args, dname+'_'+adv_arg) is None):
                        err = (f'--{dname}-{adv_arg} is required when '
                               f'--{dname} is used')
                        self._parser.error(err)
        # Set the metrics based on 'all' or the default metric
        if (args.metrics is None):
            if (args.all is True):
                args.metrics = Suspicious.getNames()
            else:
                args.metrics = [self._default_metric]
        # Set the flitsr types based on 'all' or what is set
        if (args.all is True):
            if (args.types is None):
                ts = {}
                if (getattr(args, 'refiner', None) is not None):
                    ts['refiner'] = args.refiner
                if (getattr(args, 'cluster', None) is not None):
                    ts['cluster'] = args.cluster
                if (getattr(args, 'ranker', None) is not None and
                    not default_ranker_used):
                    ts['ranker'] = args.ranker
                    args.types = [Config(**ts)]
                else:
                    args.types = [Config(**ts),
                                  Config(advanced.RankerType['FLITSR'], **ts),
                                  Config(advanced.RankerType['MULTI'], **ts)]
            if (len(args.weff) == 0 and len(args.top1) == 0 and
                    len(args.perc_at_n) == 0 and len(args.prec_rec) == 0):
                args.weff = ["first", "avg", "med", "last", 2, 3, 5]
                args.perc_at_n = ["perc"]
                args.prec_rec = [('p', 1), ('p', 5), ('p', 10), ('p', "f"),
                                 ('r', 1), ('r', 5), ('r', 10), ('r', "f")]
                args.faults = ["num"]
                args.top1 = [('all', 1), ('all', 5), ('all', 10), ('one', 5)]
        elif (args.types is None):
            args.types = [Config(args.ranker, getattr(args, 'cluster', None),
                                 getattr(args, 'refiner', None))]
        if (args.cutoff_strategy and args.cutoff_strategy.startswith('basis')):
            args.sbfl = False
            args.multi = 1
        self._add_params(args)
        return self

    def get_arg_group(self, group_name) -> Dict[str, Any]:
        """
        Return the arguments for the given argument group.

        Args:
          group_name: The name of the argument group to get arguments for. For
              example, for the SBFL advanced type, the argument name would be
              "sbfl", and would return any arguments that SBFL needs.

        Returns:
          A dictionary with all the arguments in this group as key-value pairs.
          This can be used when calling a function that requires these
          arguments using python dictionary unpacking (e.g.
          ``function_to_call(**get_arg_group(...))``).
        """
        group = {}
        params = self._advanced_params[group_name.upper()]
        prefix = group_name.lower()+'_'
        for param in params:
            if (hasattr(self, prefix+param)):
                group[param] = getattr(self, prefix+param)
            elif (hasattr(self, param)):
                group[param] = getattr(self, param)
            elif (param == 'args'):  # special case for full args
                group[param] = self
            else:
                raise KeyError(f'Could not find {param} in args!')
        return group


def get_parser() -> argparse.ArgumentParser:
    """ Return the parser for ``flitsr``. Used for the documentation """
    return Args(cmd_line=True, noparse=True)._parser


if __name__ == "__main__":
    args = Args(sys.argv[1:], cmd_line=True)
    print(args)



from typing import List
from abc import ABC, abstractmethod
from flitsr.spectrum import Spectrum
from flitsr import input


class Input(ABC):
    """ An abstract spectral input type """
    def __init_subclass__(cls):
        input.register_input(cls)

    @abstractmethod
    def get_run_file_name(self, input_path: str):
        """
        Return the name of the run file that this input type determines for
        the given input string.

        Args:
          input_path: str: The name of the input file.

        Returns:
          The name of the run file.

        """
        pass

    @abstractmethod
    def read_spectrum(self, input_path: str, split_faults: bool,
                      method_level=False) -> Spectrum:
        """
        Read in the spectrum from the given input file using this input type.

        Args:
          input_path: str: The path string of the input spectrum.
          split_faults: bool: Whether to split faults using FLITSR's
                        split_faults functionality.
          method_level:  (Default value = False) Whether to read this spectrum
                        in as a method level spectrum.

        Returns:
          The spectrum that was read in.

        """
        pass

    @classmethod
    def write_spectrum(cls, spectrum: Spectrum, output_path: str):
        """
        Write the given spectrum in the format of this input type. NOTE: this
        method does not need to be implemented by an input type. If it is not
        implemented and this method is called, a NotImplementedError will be
        thrown.

        Args:
          spectrum: Spectrum: The spectrum to write.
          output_path: str: The output path location string.

        """
        raise NotImplementedError("Cannot write spectrum in"
                                  f"{cls.__name__} format")

    @classmethod
    def get_elem_separators(cls) -> List[str]:
        """
        Method to get the separators used for printing out elements when
        writing a spectrum in the format of this input type. NOTE: this method
        does not need to be implemented by an input type unless it is used in
        the implemented write_spectrum method.
        """
        raise NotImplementedError(f"Input type {cls.__name__} not supported "
                                  "for output string")

    @classmethod
    def get_type(cls) -> 'input.InputType':
        """ Return the InputType enum of this input type. """
        return input.InputType[cls.__name__.upper()]

    @staticmethod
    @abstractmethod
    def check_format(input_path: str) -> bool:
        """
        Check whether the files from the specified input_path are of the
        concrete types format. The implementation of this abstract method
        must return a boolean, where True indicates that the input is in the
        format recognized by the implementing class, and False otherwise

        Args:
          input_path: str: The input path to check for format.

        Returns:
          True if the input path refers to input in the format that this input
          type can read, or False otherwise.

        """
        pass

    @staticmethod
    def read_in(input_path: str, split_faults: bool,
                method_level=False) -> Spectrum:
        """
        Static helper method that guesses the input type of the given input
        path out of all available input types, and reads in the spectrum using
        that input type.

        Args:
          input_path: str: The spectrum input path string to read in.
          split_faults: bool: Whether to split faults using FLITSR's
                        split_faults functionality.
          method_level:  (Default value = False) Whether to read this spectrum
                        in as a method level spectrum.

        Returns:
          The spectrum that was read in.

        """
        reader = Input.get_reader(input_path)
        return reader.read_spectrum(input_path, split_faults,
                                    method_level)

    @staticmethod
    def get_reader(input_path: str) -> 'Input':
        """
        Static helper method that guesses the input type of the given input
        path out of all available input types.

        Args:
          input_path: str: The input path string to guess the input type for.

        Returns:
          The concrete input type that can read in the given input path.

        """
        for input_enum in list(input.InputType):
            input_cls = input_enum.value
            if (input_cls.check_format(input_path)):
                return input_cls()
        raise ValueError(f"Unknown input type \"{input_path}\"")



import sys
from typing import List, Optional
from flitsr.spectrum import Spectrum, Outcome
from flitsr.ranking import Ranking, Rankings
from flitsr.input import InputType


def print_flitsr_ranking(ranking: Optional[Ranking],
                         elems: List[Spectrum.Element], file=sys.stdout):
    no_ranking = False
    if (ranking is None):  # make a tempoorary Scores object
        ranking = Ranking()
        for elem in elems:
            ranking.append(elem, 0, 0)
        no_ranking = True
    for rank in ranking:
        if (no_ranking):
            print("Faulty grouping: ", "[", file=file)
        else:
            print("Faulty grouping:", rank.score, "[", file=file)
        entity = rank.entity
        for elem in entity:
            print(" ", elem, file=file)
        print("]", file=file)


def print_rankings(rankings: Rankings, csv=False, file=sys.stdout):
    for (i, ranking) in enumerate(rankings):
        if (i > 0):
            print('<', '-'*22, ' Next Ranking ', '-'*22, '>', sep='',
                  file=file)
        if (csv):
            print_csv_ranking(ranking, file=file)
        else:
            print_flitsr_ranking(ranking, rankings.elements(), file=file)


# TODO: this function does not produce nice looking output
def print_spectrum(spectrum: Spectrum):
    for test in spectrum.tests():
        print(test.name, end=": ")
        row = spectrum[test]
        for elem in row:
            if (row[elem]):
                print(elem, end=" ")
        if (test.outcome is Outcome.PASSED):
            print('+')
        else:
            print('-')


def print_csv_ranking(ranking: Ranking, file=sys.stdout):
    print("name;suspiciousness_value", file=file)
    for rank in ranking:
        entity = rank.entity
        for elem in entity:
            print(elem.output_str(type_=InputType['GZOLTAR']),
                  ';', rank.score, sep='', file=file)


def print_spectrum_csv(spectrum: Spectrum, file=sys.stdout):
    ts = [str(t.name)+' ('+('PASS' if t.outcome is Outcome.PASSED else
                            'FAIL')+')' for t in spectrum.tests()]
    print('Element', *ts, sep=',', file=file)
    # TODO: change _elements below to elements()
    for elem in spectrum._elements:
        # print(elem, end=',', file=file)
        tests = ['X' if spectrum[t][elem] else '' for t in spectrum.tests()]
        print(elem, *tests, sep=',', file=file)
